# -*- coding: utf-8 -*-
"""
Aliases for elements which could mean the same element in XForm but is represented
differently on the XLSForm.
"""
from pyxform import constants

# Aliases:
# Ideally aliases should resolve to elements in the json form schema

# select, control and settings alias keys used for parsing,
# which is why self mapped keys are necessary.

control = {
    "group": constants.GROUP,
    "lgroup": constants.REPEAT,
    "repeat": constants.REPEAT,
    "loop": constants.LOOP,
    "looped group": constants.REPEAT,
}
select = {
    "add select one prompt using": constants.SELECT_ONE,
    "add select multiple prompt using": constants.SELECT_ALL_THAT_APPLY,
    "select all that apply from": constants.SELECT_ALL_THAT_APPLY,
    "select one from": constants.SELECT_ONE,
    "select1": constants.SELECT_ONE,
    "select_one": constants.SELECT_ONE,
    "select one": constants.SELECT_ONE,
    "select_multiple": constants.SELECT_ALL_THAT_APPLY,
    "select all that apply": constants.SELECT_ALL_THAT_APPLY,
    "select_one_external": "select one external",
    "select_one_from_file": constants.SELECT_ONE,
    "select_multiple_from_file": constants.SELECT_ALL_THAT_APPLY,
    "select one from file": constants.SELECT_ONE,
    "select multiple from file": constants.SELECT_ALL_THAT_APPLY,
    "rank": constants.RANK,
}
cascading = {
    "cascading select": constants.CASCADING_SELECT,
    "cascading_select": constants.CASCADING_SELECT,
}
settings_header = {
    "form_title": constants.TITLE,
    "set form title": constants.TITLE,
    "form_id": constants.ID_STRING,
    "sms_keyword": constants.SMS_KEYWORD,
    "sms_separator": constants.SMS_SEPARATOR,
    "sms_allow_media": constants.SMS_ALLOW_MEDIA,
    "sms_date_format": constants.SMS_DATE_FORMAT,
    "sms_datetime_format": constants.SMS_DATETIME_FORMAT,
    "prefix": constants.COMPACT_PREFIX,
    "delimiter": constants.COMPACT_DELIMITER,
    "set form id": constants.ID_STRING,
    "public_key": constants.PUBLIC_KEY,
    "submission_url": constants.SUBMISSION_URL,
    "auto_send": constants.AUTO_SEND,
    "auto_delete": constants.AUTO_DELETE,
    "allow_choice_duplicates": constants.ALLOW_CHOICE_DUPLICATES,
}
# TODO: Check on bind prefix approach in json.
# Conversion dictionary from user friendly column names to meaningful values
survey_header = {
    "Label": "label",
    "Name": "name",
    "SMS Field": constants.SMS_FIELD,
    "SMS Option": constants.SMS_OPTION,
    "SMS Separator": constants.SMS_SEPARATOR,
    "SMS Allow Media": constants.SMS_ALLOW_MEDIA,
    "SMS Date Format": constants.SMS_DATE_FORMAT,
    "SMS DateTime Format": constants.SMS_DATETIME_FORMAT,
    "SMS Response": constants.SMS_RESPONSE,
    "compact_tag": "instance::odk:tag",  # used for compact representation
    "Type": "type",
    "List_name": "list_name",
    # u"repeat_count": u"jr:count",  duplicate key
    "read_only": "bind::readonly",
    "readonly": "bind::readonly",
    "relevant": "bind::relevant",
    "caption": constants.LABEL,
    "appearance": "control::appearance",  # TODO: this is also an issue
    "relevance": "bind::relevant",
    "required": "bind::required",
    "constraint": "bind::constraint",
    "constraining message": "bind::jr:constraintMsg",
    "constraint message": "bind::jr:constraintMsg",
    "constraint_message": "bind::jr:constraintMsg",
    "calculation": "bind::calculate",
    "calculate": "bind::calculate",
    "command": constants.TYPE,
    "tag": constants.NAME,
    "value": constants.NAME,
    "image": "media::image",
    "audio": "media::audio",
    "video": "media::video",
    "count": "control::jr:count",
    "repeat_count": "control::jr:count",
    "jr:count": "control::jr:count",
    "autoplay": "control::autoplay",
    "rows": "control::rows",
    # New elements that have to go into itext elements:
    "noAppErrorString": "bind::jr:noAppErrorString",
    "no_app_error_string": "bind::jr:noAppErrorString",
    "requiredMsg": "bind::jr:requiredMsg",
    "required_message": "bind::jr:requiredMsg",
    "required message": "bind::jr:requiredMsg",
    "body": "control",
    "parameters": "parameters",
}
list_header = {
    "caption": constants.LABEL,
    "list_name": constants.LIST_NAME,
    "value": constants.NAME,
    "image": "media::image",
    "audio": "media::audio",
    "video": "media::video",
}
# Note that most of the type aliasing happens in all.xls
_type = {
    "imei": "deviceid",
    "image": "photo",
    "add image prompt": "photo",
    "add photo prompt": "photo",
    "add audio prompt": "audio",
    "add video prompt": "video",
    "add file prompt": "file",
}
yes_no = {
    "yes": True,
    "Yes": True,
    "YES": True,
    "true": True,
    "True": True,
    "TRUE": True,
    "true()": True,
    "no": False,
    "No": False,
    "NO": False,
    "false": False,
    "False": False,
    "FALSE": False,
    "false()": False,
}
label_optional_types = [
    "calculate",
    "deviceid",
    "end",
    "phonenumber",
    "simserial",
    "start",
    "start-geopoint",
    "today",
    "username",
]
osm = {"osm": constants.OSM_TYPE}



# -*- coding: utf-8 -*-
"""
AuditTest - test audit question type.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class AuditTest(PyxformTestCase):
    """
    AuditTest - test audit question type.
    """

    def test_audit(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |       |
            |        | type   |   name   | label |
            |        | audit  |   audit  |       |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary"/>',
            ],
        )

    def test_audit_random_name(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |       |
            |        | type   |   name   | label |
            |        | audit  |   bobby  |       |
            """,
            errored=True,
            error__contains=["Audits must always be named 'audit.'"],
        )

    def test_audit_blank_name(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |       |
            |        | type   |   name   | label |
            |        | audit  |          |       |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary"/>',
            ],
        )

    def test_audit_blank_parameters(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |       |            |
            |        | type   |   name   | label | parameters |
            |        | audit  |          |       |            |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary"/>',
            ],
        )

    def test_audit_location_required_parameters(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                             |
            |        | type   |   name   | parameters                                  |
            |        | audit  |   audit  | location-max-age=3, location-min-interval=1 |
            """,
            errored=True,
            error__contains=[
                "'location-priority', 'location-min-interval', and 'location-max-age' are required parameters"
            ],
        )

    def test_audit_location_priority_values(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                                                    |
            |        | type   |   name   | parameters                                                         |
            |        | audit  |   audit  | location-priority=foo, location-min-interval=1, location-max-age=2 |
            """,
            errored=True,
            error__contains=[
                "location-priority must be set to no-power, low-power, balanced, or high-accuracy"
            ],
        )

    def test_audit_location_max_age_gt_min_interval(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                                                         |
            |        | type   |   name   | parameters                                                              |
            |        | audit  |   audit  | location-priority=balanced, location-min-interval=2, location-max-age=1 |
            """,
            errored=True,
            error__contains=[
                "location-max-age must be greater than or equal to location-min-interval"
            ],
        )

    def test_audit_location_min_interval_positive(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                                                          |
            |        | type   |   name   | parameters                                                               |
            |        | audit  |   audit  | location-priority=balanced, location-min-interval=-1, location-max-age=1 |
            """,
            errored=True,
            error__contains=[
                "location-min-interval must be greater than or equal to zero"
            ],
        )

    def test_audit_location(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                                                            |
            |        | type   |   name   | parameters                                                                 |
            |        | audit  |   audit  | location-priority=balanced, location-min-interval=60, location-max-age=300 |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:location-max-age="300" odk:location-min-interval="60" odk:location-priority="balanced"/>',
            ],
        )

    def test_audit_track_changes_true(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                    |
            |        | type   |   name   | parameters         |
            |        | audit  |   audit  | track-changes=true |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:track-changes="true"/>',
            ],
        )

    def test_audit_track_changes_false(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                     |
            |        | type   |   name   | parameters          |
            |        | audit  |   audit  | track-changes=false |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:track-changes="false"/>',
            ],
        )

    def test_audit_track_changes_foo(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                   |
            |        | type   |   name   | parameters        |
            |        | audit  |   audit  | track-changes=foo |
            """,
            errored=True,
            error__contains=["track-changes must be set to true or false"],
        )

    def test_audit_identify_user_foo(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                   |
            |        | type   |   name   | parameters        |
            |        | audit  |   audit  | identify-user=foo |
            """,
            errored=True,
            error__contains=["identify-user must be set to true or false"],
        )

    def test_audit_identify_user_true(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                    |
            |        | type   |   name   | parameters         |
            |        | audit  |   audit  | identify-user=true |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:identify-user="true"/>',
            ],
        )

    def test_audit_identify_user_false(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                     |
            |        | type   |   name   | parameters          |
            |        | audit  |   audit  | identify-user=false |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:identify-user="false"/>',
            ],
        )

    def test_audit_track_changes_reasons_foo(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                          |
            |        | type   |   name   | parameters               |
            |        | audit  |   audit  | track-changes-reasons=foo |
            """,
            errored=True,
            error__contains=["track-changes-reasons must be set to on-form-edit"],
        )

    def test_audit_track_changes_reasons_on_form_edit(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                          |
            |        | type   |   name   | parameters               |
            |        | audit  |   audit  | track-changes-reasons=on-form-edit |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:track-changes-reasons="on-form-edit"/>',
            ],
        )

    def test_audit_location_track_changes(self):
        self.assertPyxformXform(
            name="meta_audit",
            md="""
            | survey |        |          |                                                                                                |
            |        | type   |   name   | parameters                                                                                     |
            |        | audit  |   audit  | location-priority=balanced, track-changes=true, location-min-interval=60, location-max-age=300 |
            """,
            xml__contains=[
                "<meta>",
                "<audit/>",
                "</meta>",
                '<bind nodeset="/meta_audit/meta/audit" type="binary" odk:location-max-age="300" odk:location-min-interval="60" odk:location-priority="balanced" odk:track-changes="true"/>',
            ],
        )



# -*- coding: utf-8 -*-
"""
Some tests for the new (v0.9) spec is properly implemented.
"""
import codecs
import os

import unittest2 as unittest

import pyxform
from pyxform.errors import PyXFormError
from pyxform.tests.utils import XFormTestCase

DIR = os.path.dirname(__file__)


class MainTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "xlsform_spec_test.xlsx"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="xlsform_spec_test", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class FlatXlsformTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "flat_xlsform_test.xlsx"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="flat_xlsform_test", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class TestNewWidgets(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "widgets.xls"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="widgets", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class CalculateWithoutCalculationTest(unittest.TestCase):
    """
    Just checks that calculate field without calculation raises a PyXFormError.
    """

    def runTest(self):
        filename = "calculate_without_calculation.xls"
        path_to_excel_file = os.path.join(DIR, "example_xls", filename)
        self.assertRaises(
            PyXFormError, pyxform.xls2json.parse_file_to_json, path_to_excel_file
        )


class PullDataTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "pull_data.xlsx"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="pull_data", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)

        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())

        # cleanup
        os.remove(self.output_path)


class SeachAndSelectTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "search_and_select.xlsx"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="search_and_select", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)

        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())

        # cleanup
        os.remove(self.output_path)


class DefaultSurveySheetTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "survey_no_name.xlsx"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)

        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Test randomize itemsets.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class RandomizeItemsetsTest(PyxformTestCase):
    def test_randomized_select_one(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                    |         |       |                |
            |        | type               | name    | label | parameters     |
            |        | select_one choices | select  | Select| randomize=true |
            | choices|                    |         |       |                |
            |        | list_name          | name    | label |                |
            |        | choices            | a       | opt_a |                |
            |        | choices            | b       | opt_b |                |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item)\">"
            ],
        )

    def test_randomized_seeded_select_one(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                    |         |       |                         |
            |        | type               | name    | label | parameters              |
            |        | select_one choices | select  | Select| randomize=true, seed=42 |
            | choices|                    |         |       |                         |
            |        | list_name          | name    | label |                         |
            |        | choices            | a       | opt_a |                         |
            |        | choices            | b       | opt_b |                         |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item, 42)\">"
            ],
        )

    def test_randomized_seeded_select_one_nameset_seed(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                    |         |       |                              |                                |
            |        | type               | name    | label | parameters                   | calculation                    |
            |        | calculate          | seed    |       |                              | once(decimal-date-time(now())) |
            |        | select_one choices | select  | Select| randomize=true,seed=${seed}  |                                |
            | choices|                    |         |       |                              |                                |
            |        | list_name          | name    | label |                              |                                |
            |        | choices            | a       | opt_a |                              |                                |
            |        | choices            | b       | opt_b |                              |                                |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item, /data/seed)\">"
            ],
        )

    def test_randomized_seeded_filtered_select_one(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                    |         |       |                         |               |
            |        | type               | name    | label | parameters              | choice_filter |
            |        | select_one choices | select  | Select| randomize=true, seed=42 | name='a'      |
            | choices|                    |         |       |                         |               |
            |        | list_name          | name    | label |                         |               |
            |        | choices            | a       | opt_a |                         |               |
            |        | choices            | b       | opt_b |                         |               |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item[name='a'], 42)\">"
            ],
        )

    def test_randomized_select_multiple(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                         |         |       |                |
            |        | type                    | name    | label | parameters     |
            |        | select_multiple choices | select  | Select| randomize=true |
            | choices|                         |         |       |                |
            |        | list_name               | name    | label |                |
            |        | choices                 | a       | opt_a |                |
            |        | choices                 | b       | opt_b |                |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item)\">"
            ],
        )

    def test_randomized_seeded_select_multiple(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                         |         |       |                         |
            |        | type                    | name    | label | parameters              |
            |        | select_multiple choices | select  | Select| randomize=true, seed=42 |
            | choices|                         |         |       |                         |
            |        | list_name               | name    | label |                         |
            |        | choices                 | a       | opt_a |                         |
            |        | choices                 | b       | opt_b |                         |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('choices')/root/item, 42)\">"
            ],
        )

    def test_randomized_external_xml_instance(self):
        self.assertPyxformXform(
            name="ecsv",
            md="""
            | survey |                                              |                |                |                |
            |        | type                                         | name           | label          | parameters     |
            |        | select_one_from_file cities.xml              | city           | City           | randomize=true |

            """,
            xml__contains=[
                "<itemset nodeset=\"randomize(instance('cities')/root/item)\">"
            ],
        )

    def test_randomized_select_one_bad_param(self):
        self.assertPyxformXform(
            name="data",
            errored="true",
            md="""
            | survey |                    |         |       |                |
            |        | type               | name    | label | parameters     |
            |        | select_one choices | select  | Select| step=10        |
            | choices|                    |         |       |                |
            |        | list_name          | name    | label |                |
            |        | choices            | a       | opt_a |                |
            |        | choices            | b       | opt_b |                |

            """,
            error__contains=[
                "Accepted parameters are 'randomize, seed': 'step' is an invalid parameter."
            ],
        )

    def test_randomized_select_one_bad_randomize(self):
        self.assertPyxformXform(
            name="data",
            errored="true",
            md="""
            | survey |                    |         |       |                  |
            |        | type               | name    | label | parameters       |
            |        | select_one choices | select  | Select| randomize=ukanga |
            | choices|                    |         |       |                  |
            |        | list_name          | name    | label |                  |
            |        | choices            | a       | opt_a |                  |
            |        | choices            | b       | opt_b |                  |

            """,
            error__contains=[
                "randomize must be set to true or false: 'ukanga' is an invalid value"
            ],
        )

    def test_randomized_select_one_bad_seed(self):
        self.assertPyxformXform(
            name="data",
            errored="true",
            md="""
            | survey |                    |         |       |                             |
            |        | type               | name    | label | parameters                  |
            |        | select_one choices | select  | Select| randomize=true, seed=ukanga |
            | choices|                    |         |       |                             |
            |        | list_name          | name    | label |                             |
            |        | choices            | a       | opt_a |                             |
            |        | choices            | b       | opt_b |                             |

            """,
            error__contains=[
                "seed value must be a number or a reference to another field."
            ],
        )

    def test_randomized_select_one_seed_without_randomize(self):
        self.assertPyxformXform(
            name="data",
            errored="true",
            md="""
            | survey |                    |         |       |                  |
            |        | type               | name    | label | parameters       |
            |        | select_one choices | select  | Select| seed=42          |
            | choices|                    |         |       |                  |
            |        | list_name          | name    | label |                  |
            |        | choices            | a       | opt_a |                  |
            |        | choices            | b       | opt_b |                  |

            """,
            error__contains=["Parameters must include randomize=true to use a seed."],
        )



import os
from pyxform.tests.utils import DIR as TESTS_DIR
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase
from pyxform.xls2xform import xls2xform_convert
from pyxform.xls2json_backends import xls_to_dict


# Common XLSForms used in below TestCases
CHOICES = """
| survey   |               |           |       |
|          | type          | name      | label |
|          | select_one l1 | q1        | Q1    |
| {name}   |               |           |       |
|          | list_name     | name      | label |
|          | l1            | 1         | C1    |
"""
# Doubled braces ${{}} here because it's used as a format string.
EXTERNAL_CHOICES = """
| survey |                        |           |       |               |
|        | type                   | name      | label | choice_filter |
|        | text                   | q1        | Q1    |               |
|        | select_one_external l1 | q2        | Q2    | q1=${{q1}}    |
| {name} |                        |           |       |               |
|        | list_name              | name      | q1    |               |
|        | l1                     | 1         | 1     |               |
|        | l1                     | 2         | 2     |               |
"""
SETTINGS = """
| survey   |           |           |       |
|          | type      | name      | label |
|          | text      | q1        | Q1    |
| {name}   |           |           |       |
|          | id_string | title     |       |
|          | my_id     | My Survey |       |
"""
SURVEY = """
| {name}   |           |           |       |
|          | type      | name      | label |
|          | text      | q1        | Q1    |
"""


class TestXLS2JSONSheetNameHeuristics(PyxformTestCase):

    err_similar_found = "the following sheets with similar names were found"
    err_survey_required = "You must have a sheet named 'survey'."
    err_choices_required = "There should be a choices sheet in this xlsform."
    err_ext_choices_required = (
        "There should be an external_choices sheet in this xlsform."
    )

    def test_workbook_to_json__case_insensitive__choices(self):
        """Should not warn/error if optional sheets are not lowercase."""
        test_names = ("choices", "Choices", "CHOICES")
        for n in test_names:
            self.assertPyxformXform(
                name="test", md=CHOICES.format(name=n), errored=False, warnings_count=0,
            )

    def test_workbook_to_json__case_insensitive__external_choices(self):
        """Should not warn/error if optional sheets are not lowercase."""
        test_names = ("external_choices", "External_Choices", "EXTERNAL_CHOICES")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=EXTERNAL_CHOICES.format(name=n),
                errored=False,
                warnings_count=0,
            )

    def test_workbook_to_json__case_insensitive__settings(self):
        """Should not warn/error if optional sheets are not lowercase."""
        test_names = ("settings", "Settings", "SETTINGS")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SETTINGS.format(name=n),
                errored=False,
                warnings_count=0,
            )

    def test_workbook_to_json__case_insensitive__survey(self):
        """Should not warn/error if the survey sheet is not lowercase."""
        test_names = ("survey", "Survey", "SURVEY")
        for n in test_names:
            self.assertPyxformXform(
                name="test", md=SURVEY.format(name=n), errored=False, warnings_count=0,
            )

    def test_workbook_to_json__ignore_prefixed_name__choices(self):
        """Should ignore sheet name for spelling if prefixed with underscore."""
        test_names = ("_choice", "_chioces", "_choics")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=CHOICES.format(name=n),
                errored=True,
                error__contains=[self.err_choices_required],
                error__not_contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__ignore_prefixed_name__external_choices(self):
        """Should ignore sheet name for spelling if prefixed with underscore."""
        test_names = ("_external_choice", "_extrenal_choices", "_externa_choics")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=EXTERNAL_CHOICES.format(name=n),
                errored=True,
                error__contains=[self.err_ext_choices_required],
                error__not_contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__ignore_prefixed_name__settings(self):
        """Should ignore sheet name for spelling if prefixed with underscore."""
        test_names = ("_setting", "_stetings", "_setings")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SETTINGS.format(name=n),
                errored=False,
                warnings_count=0,
            )

    def test_workbook_to_json__ignore_prefixed_name__survey(self):
        """Should ignore sheet name for spelling if prefixed with underscore."""
        test_names = ("_surveys", "_surve", "_sruvey")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SURVEY.format(name=n),
                errored=True,
                error__contains=[self.err_survey_required],
                error__not_contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__misspelled_found__choices(self):
        """Should mention misspellings if similar sheet names found."""
        test_names = ("choice", "chioces", "choics")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=CHOICES.format(name=n),
                errored=True,
                error__contains=[
                    self.err_choices_required,
                    self.err_similar_found,
                    "'{}'".format(n),
                ],
            )

    def test_workbook_to_json__misspelled_found__choices_exists(self):
        """Should not mention misspellings if the sheet exists."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |               |           |       |
            |          | type          | name      | label |
            |          | select_one l1 | q1        | Q1    |
            | choices  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | chioces  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_workbook_to_json__misspelled_found__choices_multiple(self):
        """Should mention misspellings if similar sheet names found."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |               |           |       |
            |          | type          | name      | label |
            |          | select_one l1 | q1        | Q1    |
            | choice   |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | chioces  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            """,
            errored=True,
            error__contains=[
                self.err_choices_required,
                self.err_similar_found,
                "'choice'",
                "'chioces'",
            ],
        )

    def test_workbook_to_json__misspelled_found__external_choices(self):
        """Should mention misspellings if similar sheet names found."""
        test_names = ("external_choice", "extrenal_choices", "externa_choics")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=EXTERNAL_CHOICES.format(name=n),
                errored=True,
                error__contains=[
                    self.err_ext_choices_required,
                    self.err_similar_found,
                    "'{}'".format(n),
                ],
            )

    def test_workbook_to_json__misspelled_found__external_choices_exists(self):
        """Should not mention misspellings if the sheet exists."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | text                   | q1        | Q1    |               |
            |          | select_one_external l1 | q2        | Q2    | q1=${q1}      |
            | external_choices |                |           |       |               |
            |          | list_name              | name      | q1    |               |
            |          | l1                     | 1         | 1     |               |
            |          | l1                     | 2         | 2     |               |
            | extrenal_choices |                |           |       |               |
            |          | list_name              | name      | q1    |               |
            |          | l1                     | 1         | 1     |               |
            |          | l1                     | 2         | 2     |               |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_workbook_to_json__misspelled_found__external_choices_multiple(self):
        """Should mention misspellings if similar sheet names found."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | text                   | q1        | Q1    |               |
            |          | select_one_external l1 | q2        | Q2    | q1=${q1}      |
            | external_choice |                 |           |       |               |
            |          | list_name              | name      | q1    |               |
            |          | l1                     | 1         | 1     |               |
            |          | l1                     | 2         | 2     |               |
            | extrenal_choices |                |           |       |               |
            |          | list_name              | name      | q1    |               |
            |          | l1                     | 1         | 1     |               |
            |          | l1                     | 2         | 2     |               |
            """,
            errored=True,
            error__contains=[
                self.err_ext_choices_required,
                self.err_similar_found,
                "'external_choice'",
                "'extrenal_choices'",
            ],
        )

    def test_workbook_to_json__misspelled_found__settings(self):
        """Should mention misspellings if similar sheet names found."""
        test_names = ("setting", "stetings", "setings")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SETTINGS.format(name=n),
                errored=False,
                warnings__contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__misspelled_found__settings_exists(self):
        """Should not mention misspellings if the sheet exists."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |           |           |       |
            |          | type      | name      | label |
            |          | text      | q1        | Q1    |
            | settings |           |           |       |
            |          | id_string | title     |       |
            |          | my_id     | My Survey |       |
            | stetings |           |           |       |
            |          | id_string | title     |       |
            |          | my_id     | My Survey |       |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_workbook_to_json__misspelled_found__settings_multiple(self):
        """Should mention misspellings if similar sheet names found."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |           |           |       |
            |          | type      | name      | label |
            |          | text      | q1        | Q1    |
            | setting  |           |           |       |
            |          | id_string | title     |       |
            |          | my_id     | My Survey |       |
            | stetings |           |           |       |
            |          | id_string | title     |       |
            |          | my_id     | My Survey |       |
            """,
            errored=False,
            warnings__contains=[self.err_similar_found, "'setting'", "'stetings'"],
        )

    def test_workbook_to_json__misspelled_found__survey(self):
        """Should mention misspellings if similar sheet names found."""
        test_names = ("surveys", "surve", "sruvey")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SURVEY.format(name=n),
                errored=True,
                error__contains=[
                    self.err_survey_required,
                    self.err_similar_found,
                    "'{}'".format(n),
                ],
            )

    def test_workbook_to_json__misspelled_found__survey_exists(self):
        """Should not mention misspellings if the sheet exists."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey  |           |           |       |
            |         | type      | name      | label |
            |         | text      | q1        | Q1    |
            | surve   |           |           |       |
            |         | type      | name      | label |
            |         | text      | q1        | Q1    |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_workbook_to_json__misspelled_found__survey_multiple(self):
        """Should mention misspellings if similar sheet names found."""
        self.assertPyxformXform(
            name="test",
            md="""
            | surveys |           |           |       |
            |         | type      | name      | label |
            |         | text      | q1        | Q1    |
            | Surve   |           |           |       |
            |         | type      | name      | label |
            |         | text      | q1        | Q1    |
            """,
            errored=True,
            error__contains=[
                self.err_survey_required,
                self.err_similar_found,
                "'surveys'",
                "'surve'",
            ],
        )

    def test_workbook_to_json__misspelled_not_found__choices(self):
        """Should not mention misspellings for dissimilar sheet names."""
        test_names = ("cho", "ices", "choose")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=CHOICES.format(name=n),
                errored=True,
                error__not_contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__misspelled_not_found__external_choices(self):
        """Should not mention misspellings for dissimilar sheet names."""
        test_names = ("external", "choices", "eternal_choosey")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=EXTERNAL_CHOICES.format(name=n),
                errored=True,
                error__not_contains=[self.err_similar_found, "'{}'".format(n)],
            )

    def test_workbook_to_json__misspelled_not_found__settings(self):
        """Should not mention misspellings for dissimilar sheet names."""
        test_names = ("hams", "spetltigs", "stetinsg")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SETTINGS.format(name=n),
                errored=False,
                warnings_count=0,
            )

    def test_workbook_to_json__misspelled_not_found__survey(self):
        """Should not mention misspellings for dissimilar sheet names."""
        test_names = ("hams", "suVVve", "settings")
        for n in test_names:
            self.assertPyxformXform(
                name="test",
                md=SURVEY.format(name=n),
                errored=True,
                error__not_contains=[self.err_similar_found],
            )

    def test_workbook_to_json__multiple_misspellings__all_ok(self):
        """Should not mention misspellings for complete example with correct spelling."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | select_one l1          | q1        | Q1    |               |
            |          | select_one_external l2 | q2        | Q2    | q1=${q1}      |
            | choices  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | external_choices |               |           |       |
            |                  | list_name     | name      | q1    |
            |                  | l2            | 1         | 1     |
            |                  | l2            | 2         | 2     |
            | settings |               |           |       |
            |          | id_string     | title     |
            |          | my_id         | My Survey |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_workbook_to_json__multiple_misspellings__survey(self):
        """Should mention misspellings in processing order (su, se, ch, ex)."""
        self.assertPyxformXform(
            name="test",
            md="""
            | surveys  |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | select_one l1          | q1        | Q1    |               |
            |          | select_one_external l2 | q2        | Q2    | q1=${q1}      |
            | chooses  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | external_choyces |               |           |       |
            |                  | list_name     | name      | q1    |
            |                  | l2            | 1         | 1     |
            |                  | l2            | 2         | 2     |
            | settyngs |               |           |       |
            |          | id_string     | title     |
            |          | my_id         | My Survey |
            """,
            errored=True,
            warnings__not_contains=[self.err_similar_found, "'settyngs'",],
            error__contains=[
                self.err_survey_required,
                self.err_similar_found,
                "'surveys'",
            ],
            error__not_contains=[
                self.err_choices_required,
                "'chooses'",
                self.err_ext_choices_required,
                "'external_choyces'",
            ],
        )

    def test_workbook_to_json__multiple_misspellings__choices(self):
        """Should mention misspellings in processing order (su, se, ch, ex)."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | select_one l1          | q1        | Q1    |               |
            |          | select_one_external l2 | q2        | Q2    | q1=${q1}      |
            | chooses  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | external_choyces |               |           |       |
            |                  | list_name     | name      | q1    |
            |                  | l2            | 1         | 1     |
            |                  | l2            | 2         | 2     |
            | settings |               |           |       |
            |          | id_string     | title     |
            |          | my_id         | My Survey |
            """,
            errored=True,
            warnings__not_contains=[self.err_similar_found, "'settyngs'",],
            error__contains=[self.err_choices_required, "'chooses'",],
            error__not_contains=[
                self.err_survey_required,
                "'survey'",
                # Not raised because the "select_one l1, q1" is checked first.
                self.err_ext_choices_required,
                "'external_choyces'",
            ],
        )

    def test_workbook_to_json__multiple_misspellings__external_choices(self):
        """Should mention misspellings in processing order (su, se, ch, ex)."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | select_one l1          | q1        | Q1    |               |
            |          | select_one_external l2 | q2        | Q2    | q1=${q1}      |
            | choices  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | external_choyces |               |           |       |
            |                  | list_name     | name      | q1    |
            |                  | l2            | 1         | 1     |
            |                  | l2            | 2         | 2     |
            | settings |               |           |       |
            |          | id_string     | title     |
            |          | my_id         | My Survey |
            """,
            errored=True,
            warnings__not_contains=[self.err_similar_found, "'settyngs'",],
            error__contains=[self.err_ext_choices_required, "'external_choyces'",],
            error__not_contains=[
                self.err_survey_required,
                "'survey'",
                self.err_choices_required,
                "'chooses'",
            ],
        )

    def test_workbook_to_json__multiple_misspellings__settings(self):
        """Should mention misspellings in processing order (su, se, ch, ex)."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |                        |           |       |               |
            |          | type                   | name      | label | choice_filter |
            |          | select_one l1          | q1        | Q1    |               |
            |          | select_one_external l2 | q2        | Q2    | q1=${q1}      |
            | chooses  |               |           |       |
            |          | list_name     | name      | label |
            |          | l1            | 1         | C1    |
            | external_choyces |               |           |       |
            |                  | list_name     | name      | q1    |
            |                  | l2            | 1         | 1     |
            |                  | l2            | 2         | 2     |
            | settyngs |               |           |       |
            |          | id_string     | title     |
            |          | my_id         | My Survey |
            """,
            errored=True,
            warnings__contains=[self.err_similar_found, "'settyngs'",],
            error__contains=[self.err_choices_required, "'chooses'",],
            error__not_contains=[
                self.err_survey_required,
                "'survey'",
                # Not raised because the "select_one l1, q1" is checked first.
                self.err_ext_choices_required,
                "'external_choyces",
            ],
        )

    def test_workbook_to_json__optional_sheets_ok(self):
        """Should not warn when valid optional sheet names are provided."""
        self.assertPyxformXform(
            name="test",
            md="""
            | survey   |           |           |       |
            |          | type      | name      | label |
            |          | text      | q1        | Q1    |
            | settings |           |           |       |
            |          | id_string | title     |       |
            |          | my_id     | My Survey |       |
            | choices  |           |           |       |
            |          | list_name | name      | label |
            |          | l1        | c1        | One   |
            """,
            errored=False,
            warnings_count=0,
        )

    def test_xls2xform_convert__e2e_with_settings_misspelling(self):
        """Should warn about settings misspelling when running full pipeline."""
        file_name = "extra_sheet_names"
        warnings = xls2xform_convert(
            xlsform_path=os.path.join(TESTS_DIR, "example_xls", file_name + ".xlsx"),
            xform_path=os.path.join(TESTS_DIR, "test_output", file_name + ".xml"),
            validate=False,
            pretty_print=False,
            enketo=False,
        )
        expected = (
            "When looking for a sheet named 'settings', the following sheets "
            "with similar names were found: 'stettings'"
        )
        self.assertIn(expected, "\n".join(warnings))

    def test_xls_to_dict__extra_sheet_names_are_returned_by_parser(self):
        """Should return all sheet names so that later steps can do spellcheck."""
        d = xls_to_dict(
            os.path.join(TESTS_DIR, "example_xls", "extra_sheet_names.xlsx")
        )
        self.assertIn("survey", d)
        self.assertIn("my_sheet", d)
        self.assertIn("stettings", d)
        self.assertIn("choices", d)



# -*- coding: utf-8 -*-
"""
Test handling setvalue of 'trigger' column in forms
"""

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class TriggerSetvalueTests(PyxformTestCase):
    def test_trigger_reference_to_nonexistent_node_gives_error(self):
        self.assertPyxformXform(
            name="trigger-missing-ref",
            md="""
            | survey |          |      |             |             |         |
            |        | type     | name | label       | calculation | trigger |
            |        | dateTime | b    |             | now()       | ${a}    |
            """,
            errored=True,
            error__contains=[
                "There has been a problem trying to replace ${a} with the XPath to the survey element named 'a'. There is no survey element with this name."
            ],
        )

    def test_trigger_with_something_other_than_node_ref_gives_error(self):
        self.assertPyxformXform(
            name="trigger-invalid-ref",
            md="""
            | survey |          |      |             |             |         |
            |        | type     | name | label       | calculation | trigger |
            |        | dateTime | b    |             | now()       | 6       |
            """,
            errored=True,
            error__contains=[
                "Only references to other fields are allowed in the 'trigger' column."
            ],
        )

    def test_handling_trigger_column_no_label_and_no_hint(self):
        md = """
        | survey |          |      |             |             |         |
        |        | type     | name | label       | calculation | trigger |
        |        | text     | a    | Enter text  |             |         |
        |        | dateTime | b    |             | now()       | ${a}    |
        """
        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<b/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/b" type="dateTime"/>',
                '<input ref="/trigger-column/a">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/b" value="now()"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/b" type="dateTime" calculate="now()"/>'
                '<input ref="/trigger-column/b">',
            ],
        )

    def test_handling_trigger_column_with_label_and_hint(self):
        md = """
        | survey |          |      |                    |             |         |
        |        | type     | name | label              | calculation | trigger |
        |        | text     | a    | Enter text         |             |         |
        |        | dateTime | c    | Date of diagnostic | now()       | ${a}    |
        """

        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<c/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/c" type="dateTime"/>',
                '<input ref="/trigger-column/a">',
                '<input ref="/trigger-column/c">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/c" value="now()"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/c" type="dateTime" calculate="now()"/>'
            ],
        )

    def test_handling_multiple_trigger_column(self):
        md = """
        | survey |          |      |            |             |         |        |
        |        | type     | name | label      | calculation | trigger | hint   |
        |        | text     | a    | Enter text |             |         |        |
        |        | integer  | b    |            | 1+1         | ${a}    |        |
        |        | dateTime | c    |            | now()       | ${a}    | A hint |
        """

        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<b/>", "<c/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/b" type="int"/>',
                '<bind nodeset="/trigger-column/c" type="dateTime"/>',
                '<input ref="/trigger-column/a">',
                '<input ref="/trigger-column/c">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/b" value="1+1"/>',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/c" value="now()"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/b" type="int" calculate="1+1"/>',
                '<bind nodeset="/trigger-column/c" type="dateTime" calculate="now()"/>',
            ],
        )

    def test_handling_trigger_column_with_no_calculation(self):
        md = """
        | survey |          |      |                    |             |         |
        |        | type     | name | label              | calculation | trigger |
        |        | text     | a    | Enter text         |             |         |
        |        | dateTime | d    | Date of something  |             | ${a}    |
        """

        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<d/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/d" type="dateTime"/>',
                '<input ref="/trigger-column/a">',
                '<input ref="/trigger-column/d">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/d"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/d" type="dateTime" calculate=""/>'
            ],
        )

    def test_handling_trigger_column_with_no_calculation_no_label_no_hint(self):
        md = """
        | survey |          |      |            |             |         |
        |        | type     | name | label      | calculation | trigger |
        |        | text     | a    | Enter text |             |         |
        |        | decimal  | e    |            |             | ${a}    |
        """

        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<e/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/e" type="decimal"/>',
                '<input ref="/trigger-column/a">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/e"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/e" type="decimal" calculate=""/>',
                '<input ref="/trigger-column/e">',
            ],
        )

    def test_trigger_refer_to_hidden_node_ref_gives_error(self):
        self.assertPyxformXform(
            name="trigger-invalid-ref",
            md="""
            | survey |           |        |                      |         |             |
            |        | type      | name   | label                | trigger | calculation |
            |        | calculate | one    |                      |         | 5 + 4       |
            |        | calculate | one-ts |                      | ${one}  | now()       |
            |        | note      | note   | timestamp: ${one-ts} |         |             |
            """,
            errored=True,
            error__contains=[
                "The question ${one} is not user-visible so it can't be used as a calculation trigger for question ${one-ts}.",
            ],
        )

    def test_when_trigger_refers_to_calculate_with_label_error_is_shown(self):
        self.assertPyxformXform(
            name="trigger-invalid-ref",
            md="""
            | survey |           |        |                      |         |             |
            |        | type      | name   | label                | trigger | calculation |
            |        | calculate | one    | A label              |         | 5 + 4       |
            |        | calculate | one-ts |                      | ${one}  | now()       |
            """,
            errored=True,
            error__contains=[
                "The question ${one} is not user-visible so it can't be used as a calculation trigger for question ${one-ts}.",
            ],
        )

    def test_typed_calculate_cant_be_trigger(self):
        self.assertPyxformXform(
            name="trigger-invalid-ref",
            md="""
                | survey |           |        |         |             |
                |        | type      | name   | trigger | calculation |
                |        | integer   | two    |         | 1 + 1       |
                |        | integer   | two-ts | ${two}  | now()       | 
                """,
            errored=True,
            error__contains=[
                "The question ${two} is not user-visible so it can't be used as a calculation trigger for question ${two-ts}."
            ],
        )

    def test_handling_trigger_column_in_group(self):
        md = """
        | survey |             |      |                    |             |         |
        |        | type        | name | label              | calculation | trigger |
        |        | text        | a    | Enter text         |             |         |
        |        | begin_group | grp  |                    |             | ${a}    |
        |        | dateTime    | c    | Date of diagnostic | now()       | ${a}    |
        |        | end_group   |      |                    |             |         |
        """

        self.assertPyxformXform(
            md=md,
            name="trigger-column",
            id_string="id",
            model__contains=["<a/>", "<grp>", "<c/>",],
            xml__contains=[
                '<bind nodeset="/trigger-column/grp/c" type="dateTime"/>',
                '<input ref="/trigger-column/a">',
                '<group ref="/trigger-column/grp">',
                '<input ref="/trigger-column/grp/c">',
                '<setvalue event="xforms-value-changed" ref="/trigger-column/grp/c" value="now()"/>',
            ],
            xml__excludes=[
                '<bind nodeset="/trigger-column/c" type="dateTime" calculate="now()"/>',
            ],
        )

    def test_calculation_with_trigger_column_should_have_expanded_xpath(self):
        self.assertPyxformXform(
            name="trigger-column",
            md="""
            | survey |          |      |             |                         |         |
            |        | type     | name | label       | calculation             | trigger |
            |        | dateTime | a    | A date      |                         |         |
            |        | integer  | b    |             | decimal-date-time(${a}) | ${a}    |
            """,
            xml__contains=[
                '<setvalue event="xforms-value-changed" ref="/trigger-column/b" value="decimal-date-time( /trigger-column/a )"/>'
            ],
        )

    def test_trigger_with_trigger_of_type_select_nests_setvalue_in_select(self):
        self.assertPyxformXform(
            name="trigger-select_trigger",
            md="""
            | survey |                    |      |             |                    |         |
            |        | type               | name | label       | calculation        | trigger |
            |        | select_one choices | a    | Some choice |                    |         |
            |        | integer            | b    |             | string-length(${a})| ${a}    |
            | choices|                    |      |             |                    |         |
            |        | list_name          | name | label       |
            |        | choices            | a    | A           |
            |        | choices            | aa   | AA          |
            """,
            xml__contains=[
                '<setvalue event="xforms-value-changed" ref="/trigger-select_trigger/b" value="string-length( /trigger-select_trigger/a )"/>\n    </select1>'
            ],
        )

    def test_trigger_column_in_repeat_should_have_expanded_xpath(self):
        self.assertPyxformXform(
            name="trigger-column",
            md="""
            | survey |              |       |                        |              |         |
            |        | type         | name  | label                  | calculation  | trigger |
            |        | begin repeat | rep   |                        |              |         |
            |        | dateTime     | one   | Enter text             |              |         |
            |        | dateTime     | three | Enter text (triggered) | now()        | ${one}  |
            |        | end repeat   |       |                        |              |         |
            """,
            xml__contains=[
                '<setvalue event="xforms-value-changed" ref="/trigger-column/rep/three" value="now()"/>'
            ],
        )

    def test_trigger_column_in_repeat_should_have_expanded_xpath_in_value(self):
        self.assertPyxformXform(
            name="trigger-column",
            md="""
            | survey |              |       |                        |                        |         |
            |        | type         | name  | label                  | calculation            | trigger |
            |        | begin repeat | rep   |                        |                        |         |
            |        | dateTime     | one   | Enter text             |                        |         |
            |        | dateTime     | three | Enter text (triggered) | string-length(${one})  | ${one}  |
            |        | end repeat   |       |                        |                        |         |
            """,
            xml__contains=[
                '<setvalue event="xforms-value-changed" ref="/trigger-column/rep/three" value="string-length( ../one )"/>'
            ],
        )



# -*- coding: utf-8 -*-
"""
Test xls2json_backends util functions.
"""
from unittest import TestCase

from pyxform.tests import utils
from pyxform.xls2json_backends import convert_file_to_csv_string


class BackendUtilsTests(TestCase):
    def test_xls_to_csv(self):
        specify_other_xls = utils.path_to_text_fixture("specify_other.xls")
        converted_xls = convert_file_to_csv_string(specify_other_xls)
        specify_other_csv = utils.path_to_text_fixture("specify_other.csv")
        converted_csv = convert_file_to_csv_string(specify_other_csv)
        self.assertEqual(converted_csv, converted_xls)



# -*- coding: utf-8 -*-
"""
Test validator update cli command.
"""
import os
import platform
import shutil
import tempfile
from contextlib import contextmanager
from datetime import datetime, timedelta
from stat import S_IXGRP, S_IXUSR
from zipfile import ZipFile

from unittest2 import TestCase, skipIf

from pyxform.errors import PyXFormError
from pyxform.tests import validators
from pyxform.tests.validators.server import ThreadingServerInThread
from pyxform.utils import unicode
from pyxform.validators.updater import (
    EnketoValidateUpdater,
    _UpdateHandler,
    _UpdateInfo,
    capture_handler,
)

try:
    from zipfile import BadZipFile
except ImportError:
    from zipfile import BadZipfile as BadZipFile


TEST_PATH = validators.HERE


def install_check_ok(bin_file_path=None):
    return True


def install_check_fail(bin_file_path=None):
    return False


def get_update_info(check_ok, mod_root=None):
    """
    Get an UpdateInfo for testing use.

    :type check_ok: bool
    :type mod_root: str
    :return: _UpdateInfo
    """
    if check_ok:
        install_check = install_check_ok
    else:
        install_check = install_check_fail
    return _UpdateInfo(
        api_url="",
        repo_url="",
        validate_subfolder="",
        install_check=install_check,
        validator_basename="validate",
        mod_root=mod_root,
    )


@contextmanager
def get_temp_file():
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.close()
    try:
        yield temp_file.name
    finally:
        temp_file.close()
        if os.path.exists(temp_file.name):
            os.remove(temp_file.name)


@contextmanager
def get_temp_dir():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)


class TestTempUtils(TestCase):
    def test_get_temp_file(self):
        """Should provide a temp file that's cleared on exit."""
        with get_temp_file() as temp_file:
            self.assertTrue(os.path.exists(temp_file))
            self.assertTrue(os.path.isfile(temp_file))
        self.assertFalse(os.path.exists(temp_file))

    def test_get_temp_dir(self):
        """Should provide a temp dir that's cleared on exit."""
        with get_temp_dir() as temp_dir:
            self.assertTrue(os.path.exists(temp_dir))
            self.assertTrue(os.path.isdir(temp_dir))
        self.assertFalse(os.path.exists(temp_dir))


class TestUpdateHandler(TestCase):

    server = ThreadingServerInThread()

    @classmethod
    def setUpClass(cls):
        cls.server.start()

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()

    def setUp(self):
        self.update_info = get_update_info(check_ok=True)
        self.updater = _UpdateHandler()
        data_dir = os.path.join(TEST_PATH, "data")
        self.latest_enketo = os.path.join(data_dir, "latest_enketo.json")
        self.latest_odk = os.path.join(data_dir, "latest_odk.json")
        self.last_check = os.path.join(TEST_PATH, ".last_check")
        self.last_check_none = os.path.join(TEST_PATH, ".last_check_none")
        self.phantom_file = os.path.join(TEST_PATH, ".not_there")
        self.utc_now = datetime.utcnow()
        capture_handler.reset()
        self.zip_file = os.path.join(data_dir, "linux-ideal.zip")
        self.zip_file_ideal = os.path.join(data_dir, "linux-ideal.zip")
        self.zip_file_dupes = os.path.join(data_dir, "linux-dupes.zip")
        self.install_fake = os.path.join(data_dir, "install_fake.json")
        self.install_fake_old = os.path.join(data_dir, "install_fake_old.json")

    def test_request_latest_json(self):
        """Should return version info dict containing asset list."""
        self.update_info.api_url = "http://localhost:8000/latest_enketo.json"
        observed = self.updater._request_latest_json(url=self.update_info.api_url)
        self.assertIn("assets", observed)

    def test_check_path__raises(self):
        """Should raise an error if the path doesn't exist."""
        file_path = os.path.join(TEST_PATH, "data", "non_existent.json")
        with self.assertRaises(PyXFormError):
            self.updater._check_path(file_path=file_path)

    def test_check_path__file(self):
        """Should return True if the file path exists."""
        self.assertTrue(self.updater._check_path(file_path=self.last_check))

    def test_check_path__dir(self):
        """Should return True if the directory path exists."""
        self.assertTrue(self.updater._check_path(file_path=TEST_PATH))

    def test_read_json(self):
        """Should return version info dict containing asset list."""
        file_path = self.latest_enketo
        observed = self.updater._read_json(file_path=file_path)
        self.assertIn("assets", observed)

    def test_write_json(self):
        """Should write the supplied dict to a file."""
        with get_temp_file() as temp_file:
            self.assertEqual(0, os.path.getsize(temp_file))
            self.updater._write_json(file_path=temp_file, content={"some": "data"})
            self.assertTrue(20 <= os.path.getsize(temp_file))

    def test_read_last_check(self):
        """Should return a datetime from the last_check file."""
        file_path = self.last_check
        last_check = self.updater._read_last_check(file_path=file_path)
        self.assertTrue(isinstance(last_check, datetime))

    def test_write_last_check(self):
        """Should write the supplied datetime to a file."""
        with get_temp_file() as temp_file:
            self.assertEqual(0, os.path.getsize(temp_file))
            self.updater._write_last_check(
                file_path=temp_file, content=datetime.utcnow()
            )
            self.assertEqual(20, os.path.getsize(temp_file))

    def test_check_necessary__true_if_last_check_not_found(self):
        """Should return true if the last check file wasn't found."""
        self.update_info.last_check_path = self.phantom_file
        self.assertTrue(
            self.updater._check_necessary(
                update_info=self.update_info, utc_now=self.utc_now
            )
        )

    def test_check_necessary__true_if_latest_json_not_found(self):
        """Should return true if the latest.json file wasn't found."""
        self.update_info.last_check_path = self.last_check
        self.update_info.latest_path = self.phantom_file
        self.assertTrue(
            self.updater._check_necessary(
                update_info=self.update_info, utc_now=self.utc_now
            )
        )

    def test_check_necessary__true_if_last_check_empty(self):
        """Should return true if the last check file was empty."""
        self.update_info.last_check_path = os.path.join(TEST_PATH, ".last_check_none")
        self.update_info.latest_path = self.latest_enketo
        self.assertTrue(
            self.updater._check_necessary(
                update_info=self.update_info, utc_now=self.utc_now
            )
        )

    def test_check_necessary__true_if_last_check_too_old(self):
        """Should return true if the last check was too long ago."""
        self.update_info.last_check_path = self.last_check
        self.update_info.latest_path = self.latest_enketo
        old = self.utc_now - timedelta(minutes=45.0)
        self.assertTrue(
            self.updater._check_necessary(update_info=self.update_info, utc_now=old)
        )

    def test_check_necessary__false_last_check_very_recent(self):
        """Should return false if the last check was very recent."""
        new = self.utc_now - timedelta(minutes=10.0)

        with get_temp_file() as temp_file:
            self.updater._write_last_check(file_path=temp_file, content=new)
            self.update_info.last_check_path = temp_file
            self.update_info.latest_path = self.latest_enketo
            self.assertFalse(
                self.updater._check_necessary(
                    update_info=self.update_info, utc_now=self.utc_now
                )
            )

    def test_get_latest__if_check_necessary_true(self):
        """Should get latest from remote, rather than file."""
        self.update_info.api_url = "http://localhost:8000/latest_enketo.json"
        old = self.utc_now - timedelta(minutes=45.0)

        with get_temp_file() as temp_check, get_temp_file() as temp_json:
            self.updater._write_last_check(file_path=temp_check, content=old)
            self.update_info.last_check_path = temp_check
            self.update_info.latest_path = temp_json
            latest = self.updater._get_latest(update_info=self.update_info)
        self.assertEqual("1.0.3", latest["name"])

    def test_get_latest__if_check_necessary_false(self):
        """Should get latest from file, rather than remote."""
        self.update_info.latest_path = self.latest_odk
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_file() as temp_check:
            self.updater._write_last_check(file_path=temp_check, content=new)
            self.update_info.last_check_path = temp_check
            latest = self.updater._get_latest(update_info=self.update_info)
        self.assertEqual("ODK Validate v1.8.0", latest["name"])

    def test_list__not_installed_no_files(self):
        """Should log an info message - no installed version, no files."""
        self.update_info.installed_path = self.phantom_file
        self.update_info.latest_path = self.latest_odk

        with get_temp_file() as temp_check:
            self.updater._write_last_check(file_path=temp_check, content=self.utc_now)
            self.update_info.last_check_path = temp_check
            self.updater.list(update_info=self.update_info)
        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Installed release:\n\n- None!", info)
        self.assertIn("Files available:\n\n- None!", info)

    def test_list__not_installed_with_files(self):
        """Should log an info message - no installed version, with files."""
        self.update_info.installed_path = self.phantom_file
        self.update_info.latest_path = self.latest_enketo

        with get_temp_file() as temp_check:
            self.updater._write_last_check(file_path=temp_check, content=self.utc_now)
            self.update_info.last_check_path = temp_check
            self.updater.list(update_info=self.update_info)
        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Installed release:\n\n- None!", info)
        self.assertIn("- windows.zip", info)

    def test_list__installed_no_files(self):
        """Should log an info message - installed version, no files."""
        self.update_info.installed_path = self.latest_enketo
        self.update_info.latest_path = self.latest_odk

        with get_temp_file() as temp_check:
            self.updater._write_last_check(file_path=temp_check, content=self.utc_now)
            self.update_info.last_check_path = temp_check
            self.updater.list(update_info=self.update_info)
        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Installed release:\n\n- Tag name = 1.0.3", info)
        self.assertIn("Files available:\n\n- None!", info)

    def test_list__installed_with_files(self):
        """Should log an info message - installed version, with files."""
        self.update_info.installed_path = self.latest_enketo
        self.update_info.latest_path = self.latest_enketo

        with get_temp_file() as temp_check:
            self.updater._write_last_check(file_path=temp_check, content=self.utc_now)
            self.update_info.last_check_path = temp_check
            self.updater.list(update_info=self.update_info)
        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Installed release:\n\n- Tag name = 1.0.3", info)
        self.assertIn("- windows.zip", info)

    def test_find_download_url__no_files(self):
        """Should raise an error if no files attached to release."""
        file_name = "windows.zip"
        json_data = self.updater._read_json(file_path=self.latest_odk)

        with self.assertRaises(PyXFormError) as ctx:
            self.updater._find_download_url(
                update_info=self.update_info, json_data=json_data, file_name=file_name
            )
        self.assertIn("No files attached", unicode(ctx.exception))

    def test_find_download_url__not_found(self):
        """Should raise an error if the file was not found."""
        file_name = "windows.zip"
        json_data = self.updater._read_json(file_path=self.latest_enketo)
        json_data["assets"] = [x for x in json_data["assets"] if x["name"] != file_name]

        with self.assertRaises(PyXFormError) as ctx:
            self.updater._find_download_url(
                update_info=self.update_info, json_data=json_data, file_name=file_name
            )
        self.assertIn("No files with the name", unicode(ctx.exception))

    def test_find_download_url__duplicates(self):
        """Should raise an error if the file was found more than once."""
        file_name = "windows.zip"
        json_data = self.updater._read_json(file_path=self.latest_enketo)
        file_dicts = [x for x in json_data["assets"] if x["name"] == file_name]
        json_data["assets"].append(file_dicts[0])

        with self.assertRaises(PyXFormError) as ctx:
            self.updater._find_download_url(
                update_info=self.update_info, json_data=json_data, file_name=file_name
            )
        self.assertIn("2 files with the name", unicode(ctx.exception))

    def test_find_download_url__ok(self):
        """Should return the url for the matching file name."""
        file_name = "windows.zip"
        json_data = self.updater._read_json(file_path=self.latest_enketo)
        expected = (
            "https://github.com/enketo/enketo-validate/releases/"
            "download/1.0.3/windows.zip"
        )

        observed = self.updater._find_download_url(
            update_info=self.update_info, json_data=json_data, file_name=file_name
        )
        self.assertEqual(expected, observed)

    def test_download_file(self):
        """Should download the file from the url to the the target path."""
        self.update_info.api_url = "http://localhost:8000/.small_file"
        with get_temp_file() as temp_file:
            self.assertEqual(0, os.path.getsize(temp_file))
            self.updater._download_file(
                url=self.update_info.api_url, file_path=temp_file
            )
            self.assertEqual(13, os.path.getsize(temp_file))

    def test_get_bin_paths__ok(self):
        """Should return the path mappings."""
        file_path = os.path.join(TEST_PATH, "linux.zip")
        observed = self.updater._get_bin_paths(
            update_info=self.update_info, file_path=file_path
        )
        self.assertEqual(3, len(observed))

    def test_get_bin_paths__unsupported_raises(self):
        """Should raise an error if a mapping for the file name isn't found."""
        file_path = self.last_check = os.path.join(TEST_PATH, "bacon.zip")
        with self.assertRaises(PyXFormError) as ctx:
            self.updater._get_bin_paths(
                update_info=self.update_info, file_path=file_path
            )
        self.assertIn("Did not find", unicode(ctx.exception))

    def test_unzip_find_zip_jobs__ok_real_current(self):
        """Should return a list of zip jobs same length as search."""
        with get_temp_dir() as temp_dir, ZipFile(self.zip_file, mode="r") as zip_file:
            bin_paths = self.updater._get_bin_paths(
                update_info=self.update_info, file_path=self.zip_file
            )
            jobs = self.updater._unzip_find_jobs(
                open_zip_file=zip_file, bin_paths=bin_paths, out_path=temp_dir
            )
        self.assertEqual(3, len(jobs.keys()))
        self.assertTrue(list(jobs.keys())[0].startswith(temp_dir))

    def test_unzip_find_zip_jobs__ok_real_ideal(self):
        """Should return a list of zip jobs same length as search."""
        with get_temp_dir() as temp_dir, ZipFile(
            self.zip_file_ideal, mode="r"
        ) as zip_file:
            bin_paths = self.updater._get_bin_paths(
                update_info=self.update_info, file_path=self.zip_file_ideal
            )
            jobs = self.updater._unzip_find_jobs(
                open_zip_file=zip_file, bin_paths=bin_paths, out_path=temp_dir
            )
        self.assertEqual(3, len(jobs.keys()))
        self.assertTrue(list(jobs.keys())[0].startswith(temp_dir))

    def test_unzip_find_zip_jobs__ok_real_dupes(self):
        """Should return a list of zip jobs same length as search."""
        with get_temp_dir() as temp_dir, ZipFile(
            self.zip_file_dupes, mode="r"
        ) as zip_file:
            bin_paths = self.updater._get_bin_paths(
                update_info=self.update_info, file_path=self.zip_file_dupes
            )
            jobs = self.updater._unzip_find_jobs(
                open_zip_file=zip_file, bin_paths=bin_paths, out_path=temp_dir
            )
        self.assertEqual(3, len(jobs.keys()))
        self.assertTrue(list(jobs.keys())[0].startswith(temp_dir))

    def test_unzip_find_zip_jobs__not_found_raises(self):
        """Should raise an error if zip jobs isn't same length as search."""
        bin_paths = [(".non_existent", ".non_existent")]

        with get_temp_dir() as temp_dir, ZipFile(
            self.zip_file, mode="r"
        ) as zip_file, self.assertRaises(PyXFormError) as ctx:
            self.updater._unzip_find_jobs(
                open_zip_file=zip_file, bin_paths=bin_paths, out_path=temp_dir
            )
        self.assertIn("1 zip job files, found: 0", unicode(ctx.exception))

    def test_unzip_extract_file__ok(self):
        """Should extract the specified item to the target output path."""
        with get_temp_dir() as temp_dir, ZipFile(self.zip_file, mode="r") as zip_file:
            zip_item = zip_file.infolist()[0]
            file_out_path = os.path.join(temp_dir, "validate")
            self.updater._unzip_extract_file(
                open_zip_file=zip_file, zip_item=zip_item, file_out_path=file_out_path
            )
            self.assertTrue(os.path.exists(file_out_path))

    def test_unzip_extract_file__bad_crc_raises(self):
        """Should raise an error if the zip file CRC doesn't match."""
        with get_temp_dir() as temp_dir, ZipFile(
            self.zip_file, mode="r"
        ) as zip_file, self.assertRaises(BadZipFile) as ctx:
            zip_item = [
                x for x in zip_file.infolist() if x.filename.endswith("validate")
            ][0]
            zip_item.CRC = 12345
            file_out_path = os.path.join(temp_dir, "validate")
            self.updater._unzip_extract_file(
                open_zip_file=zip_file, zip_item=zip_item, file_out_path=file_out_path
            )
        self.assertIn("Bad CRC-32 for file", unicode(ctx.exception))

    def test_unzip(self):
        """Should unzip the file to the locations in the bin_path map."""
        with get_temp_dir() as temp_dir:
            self.updater._unzip(
                update_info=self.update_info, file_path=self.zip_file, out_path=temp_dir
            )
            dir_list = [
                os.path.join(r, f) for r, _, fs in os.walk(temp_dir) for f in fs
            ]
            self.assertEqual(3, len(dir_list))

    def test_install__ok(self):
        """Should install the latest release and return it's info dict."""
        self.update_info.latest_path = self.install_fake
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_file() as temp_check, get_temp_dir() as temp_dir:
            self.updater._write_last_check(file_path=temp_check, content=new)
            self.update_info.last_check_path = temp_check
            self.update_info.bin_new_path = temp_dir
            installed = self.updater._install(
                update_info=self.update_info, file_name="linux.zip"
            )
            dir_list = [
                os.path.join(r, f) for r, _, fs in os.walk(temp_dir) for f in fs
            ]
            self.assertEqual(5, len(dir_list))

        latest = self.updater._read_json(file_path=self.install_fake)
        self.assertDictEqual(latest, installed)

    @skipIf(platform.system() == "Windows", "Exec bits can't be set on Windows")
    def test_install__add_executable_mode(self):
        """Should add executable mode to the new bin file's modes."""
        self.update_info.latest_path = self.install_fake
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_file() as temp_check, get_temp_dir() as temp_dir:
            self.updater._write_last_check(file_path=temp_check, content=new)
            self.update_info.last_check_path = temp_check
            self.update_info.bin_new_path = temp_dir
            self.updater._install(update_info=self.update_info, file_name="linux.zip")
            bin_new = os.path.join(temp_dir, self.update_info.validator_basename)
            bin_new_stat_mode = os.stat(bin_new).st_mode
            self.assertEqual(bin_new_stat_mode & S_IXUSR, S_IXUSR)
            self.assertEqual(bin_new_stat_mode & S_IXGRP, S_IXGRP)

    def test_replace_old_bin_path(self):
        """Should delete the old bin path and move new into it's place."""
        with get_temp_dir() as installed, get_temp_dir() as staging:
            self.update_info.bin_path = installed
            self.update_info.bin_new_path = staging
            lcp = os.path.join(staging, ".last_check")
            self.updater._write_last_check(file_path=lcp, content=self.utc_now)
            self.updater._replace_old_bin_path(update_info=self.update_info)

            self.assertFalse(os.path.exists(staging))
            self.assertEqual(1, len(os.listdir(installed)))
            self.assertTrue(os.path.exists(installed))

    def test_update__not_installed__ok(self):
        """Should install and show a message with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root:
            update_info = get_update_info(check_ok=True, mod_root=mod_root)
            update_info.latest_path = self.install_fake
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            expected_path = os.path.join(update_info.bin_path, "validate")
            self.assertFalse(os.path.exists(expected_path))
            self.updater.update(update_info=update_info, file_name="linux.zip")
            self.assertTrue(os.path.exists(expected_path))

        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Update success!", info)

    def test_update__not_installed__fail__install_check(self):
        """Should stop install and raise an error with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root, self.assertRaises(PyXFormError) as ctx:
            update_info = get_update_info(check_ok=False, mod_root=mod_root)
            update_info.latest_path = self.install_fake
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.assertFalse(os.path.exists(update_info.bin_path))
            self.updater.update(update_info=update_info, file_name="linux.zip")
            self.assertFalse(os.path.exists(update_info.bin_path))
            self.assertTrue(os.path.exists(update_info.bin_new_path))

        error = unicode(ctx.exception)
        self.assertIn("Update failed!", error)
        self.assertIn("latest release does not appear to work", error)

    def test_update__installed__ok(self):
        """Should update and show a message with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root:
            update_info = get_update_info(check_ok=True, mod_root=mod_root)
            update_info.latest_path = self.install_fake_old
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.updater.update(update_info=update_info, file_name="linux.zip")
            update_info.latest_path = self.install_fake
            self.updater.update(update_info=update_info, file_name="linux.zip")

        info = capture_handler.watcher.output["INFO"][0]
        self.assertIn("Update success!", info)
        self.assertIn("Install check of the latest release succeeded", info)

    def test_update__installed__fail__already_latest(self):
        """Should stop install and raise an error with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root, self.assertRaises(PyXFormError) as ctx:
            update_info = get_update_info(check_ok=True, mod_root=mod_root)
            update_info.latest_path = self.install_fake
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.updater.update(update_info=update_info, file_name="linux.zip")
            update_info.latest_path = self.install_fake
            self.updater.update(update_info=update_info, file_name="linux.zip")

        error = unicode(ctx.exception)
        self.assertIn("Update failed!", error)
        self.assertIn("installed release appears to be the latest", error)

    def test_update__installed__fail__install_check(self):
        """Should stop install and raise an error with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root, self.assertRaises(PyXFormError) as ctx:
            update_info = get_update_info(check_ok=False, mod_root=mod_root)
            update_info.latest_path = self.install_fake
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.updater.update(update_info=update_info, file_name="linux.zip")
            update_info.latest_path = self.install_fake
            self.updater.update(update_info=update_info, file_name="linux.zip")

            self.assertFalse(os.path.exists(update_info.bin_path))
            self.assertTrue(os.path.exists(update_info.bin_new_path))

        error = unicode(ctx.exception)
        self.assertIn("Update failed!", error)
        self.assertIn("latest release does not appear to work", error)

    def test_check__fail__not_installed(self):
        """Should raise an error if there's no installation detected."""
        self.update_info.installed_path = os.path.join(TEST_PATH, ".nothing")
        with self.assertRaises(PyXFormError) as ctx:
            self.updater.check(self.update_info)

        error = unicode(ctx.exception)
        self.assertIn("Check failed!", error)
        self.assertIn("No installed release found", error)

    def test_check__ok(self):
        """Should show a message with relevant info."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root:
            update_info = get_update_info(check_ok=True, mod_root=mod_root)
            update_info.latest_path = self.install_fake_old
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.updater.update(update_info=update_info, file_name="linux.zip")
            self.updater.check(update_info=update_info)

        info = capture_handler.watcher.output["INFO"][1]
        self.assertIn("Check success!", info)
        self.assertIn("installed release appears to work", info)

    def test_check__fail__install_check(self):
        """Should raise an error if the installation check fails."""
        new = self.utc_now - timedelta(minutes=15.0)

        with get_temp_dir() as mod_root, self.assertRaises(PyXFormError) as ctx:
            update_info = get_update_info(check_ok=True, mod_root=mod_root)
            update_info.latest_path = self.install_fake_old
            self.updater._write_last_check(
                file_path=update_info.last_check_path, content=new
            )

            self.updater.update(update_info=update_info, file_name="linux.zip")
            update_info.install_check = install_check_fail
            self.updater.check(update_info=update_info)

        error = unicode(ctx.exception)
        self.assertIn("Check failed!", error)
        self.assertIn("installed release does not appear to work", error)

    def test_enketo_validate_updater__install_check_routing_ok(self):
        """Should call the install check on the UpdateInfo instance."""
        ev = EnketoValidateUpdater()
        ev.update_info.install_check = install_check_ok
        ev.update_info.installed_path = self.install_fake
        self.assertTrue(ev.check())

    def test_enketo_validate_updater__install_check_routing_fail(self):
        """Should raise if the install check function is bogus."""
        ev = EnketoValidateUpdater()
        ev.update_info.install_check = None
        ev.update_info.installed_path = self.install_fake
        with self.assertRaises(TypeError):
            ev.check()



# -*- coding: utf-8 -*-
"""
Test file question type.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class FileWidgetTest(PyxformTestCase):
    """
    Test file widget class.
    """

    def test_file_type(self):
        """
        Test file question type.
        """
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |        |               |
            |        | type   | name   | label         |
            |        | file   | file   | Attach a file |
            """,
            xml__contains=['<upload mediatype="application/*"'],
        )



# -*- coding: utf-8 -*-
"""
Testing inlining translation when no translation is specified.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class TestSecondaryInstanceTest(PyxformTestCase):
    def test_inline_translations(self):
        """
        Dynamic choice (marked with choice filter) should inline the labels instead of using
        itext fields. There should be no itext definition in the model.
        """
        self.assertPyxformXform(
            md="""
               | survey  |                   |          |            |               |
               |         | type              | name     | label      | choice_filter |
               |         | select_one states | state    | State Name | state != ''   |
               | choices |                   |          |            |
               |         | list name         | name     | label      |
               |         | states            | option a | a          |
               |         | states            | option b | b          |
               |         | states            | option c | c          |
               """,
            name="data",
            id_string="some-id",
            model__contains=[
                "<label>a</label>",
                "<label>b</label>",
                "<label>c</label>",
            ],
            model__excludes=[
                '<text id="states-0">',
                '<text id="states-1">',
                '<text id="states-2">',
                "<itextId>states-0</itextId>",
                "<itextId>states-1</itextId>",
                "<itextId>states-2</itextId>",
            ],
            xml__contains=['<label ref="label"/>'],
            xml__excludes=['<label ref="jr:itext(itextId)"/>'],
        )

    def test_multiple_translations(self):
        """
        Dynamic choice with potential translation should generate itext fields.
        """
        self.assertPyxformXform(
            md="""
               | survey  |                   |          |            |               |
               |         | type              | name     | label      | choice_filter |
               |         | select_one states | state    | State Name | state != ''   |
               | choices |                   |          |            |
               |         | list name         | name     | label::English(en)|
               |         | states            | option a | a                 |
               |         | states            | option b | b                 |
               |         | states            | option c | c                 |
               """,
            name="data",
            id_string="some-id",
            model__contains=[
                '<text id="states-0">',
                '<text id="states-1">',
                '<text id="states-2">',
                "<itextId>states-0</itextId>",
                "<itextId>states-1</itextId>",
                "<itextId>states-2</itextId>",
            ],
            model__excludes=[
                "<label>a</label>",
                "<label>b</label>",
                "<label>c</label>",
            ],
            xml__contains=['<label ref="jr:itext(itextId)"/>'],
            xml__excludes=['<label ref="label"/>'],
        )

    def test_select_with_media_and_choice_filter_and_no_translations_generates_media(
        self,
    ):
        """
        Selects with media and choice filter should generate itext fields for the media.
        """
        xform_md = """
        | survey |                    |                 |                                 |                           |
        |        | type               | name            | label                           | choice_filter             |
        |        | select_one consent | consent         | Would you like to participate ? |                           |
        |        | select_one mood    | enumerator_mood | How are you feeling today ?     | selected(${consent}, 'y') |
        | choices |
        |         | list_name | name | label | media::image |
        |         | mood      | h    | Happy | happy.jpg    |
        |         | mood      | s    | Sad   | sad.jpg      |
        |         | consent   | y    | Yes   |              |
        |         | consent   | n    | No    |              |
        """
        self.assertPyxformXform(
            name="data",
            id_string="some-id",
            md=xform_md,
            errored=False,
            model__contains=[
                '<text id="mood-0">',
                '<text id="mood-1">',
                "<itextId>mood-0</itextId>",
                "<itextId>mood-1</itextId>",
            ],
            model__excludes=[
                '<text id="/data/enumerator_mood/h:label">',
                '<text id="/data/enumerator_mood/s:label">',
            ],
            xml__contains=['<label ref="jr:itext(itextId)"/>'],
            xml__excludes=['<label ref="label"/>', "<label>Happy</label>"],
        )

    def test_select_with_choice_filter_and_translations_generates_single_translation(
        self,
    ):
        """
        Selects with choice filter and translations should only have a single itext entry.
        """
        xform_md = """
        | survey |                    |      |       |               |
        |        | type               | name | label | choice_filter |
        |        | select_one list    | foo  | Foo   | name != ''    |
        | choices |
        |         | list_name | name | label | image | label::French |
        |         | list      | a    | A     | a.jpg | Ah            |
        |         | list      | b    | B     | b.jpg | Bé            |
        |         | list      | c    | C     | c.jpg | Cé            |
        """
        self.assertPyxformXform(
            name="data",
            id_string="some-id",
            md=xform_md,
            errored=False,
            debug=False,
            itext__contains=[
                '<text id="list-0">',
                '<text id="list-1">',
                '<text id="list-2">',
            ],
            itext__excludes=[
                '<text id="/data/foo/a:label">',
                '<text id="/data/foo/b:label">',
                '<text id="/data/foo/c:label">',
            ],
        )

    def test_select_with_dynamic_option_label__and_choice_filter__and_no_translations__generates_itext(
        self,
    ):
        """
        A select with a choice filter and no translations in which the first option label is dynamic should generate itext for choice labels.
        """
        xform_md = """
            | survey |                    |      |            |               |         |
            |        | type               | name | label      | choice_filter | default |
            |        | text               | txt  | Enter text |               | default |
            |        | select_one choices | one  | Select one | 1 < 2         |         |
            | choices |
            |         | list_name | name | label        |
            |         | choices   | one  | One - ${txt} |
            """
        self.assertPyxformXform(
            name="data",
            md=xform_md,
            debug=False,
            itext__contains=[
                '<text id="choices-0">',
                '<value> One - <output value=" /data/txt "/>',
            ],
            model__contains=["<itextId>choices-0</itextId>", "<name>one</name>",],
            xml__contains=['<label ref="jr:itext(itextId)"/>'],
            xml__excludes=['<label ref="label"/>', "<label>One - ${txt}</label>"],
        )

    def test_select_with_dynamic_option_label_for_second_choice__and_choice_filter__and_no_translations__generates_itext(
        self,
    ):
        """
        A select with a choice filter and no translations in which the second option label is dynamic should generate itext for choice labels.
        """
        xform_md = """
            | survey |                    |      |            |               |         |
            |        | type               | name | label      | choice_filter | default |
            |        | text               | txt  | Enter text |               | default |
            |        | select_one choices | one  | Select one | 1 < 2         |         |
            | choices |
            |         | list_name | name | label        |
            |         | choices   | one  | One          |
            |         | choices   | two  | Two - ${txt} |
            """
        self.assertPyxformXform(
            name="data",
            md=xform_md,
            itext__contains=[
                '<text id="choices-0">',
                "<value>One</value>",
                '<text id="choices-1">',
                '<value> Two - <output value=" /data/txt "/>',
            ],
            model__contains=[
                "<itextId>choices-0</itextId>",
                "<name>one</name>",
                "<itextId>choices-1</itextId>",
                "<name>two</name>",
            ],
            xml__contains=['<label ref="jr:itext(itextId)"/>'],
            xml__excludes=['<label ref="label"/>', "<label>One</label>"],
        )

    def test_select_with_dynamic_option_label__and_choice_filter__and_no_translations__maintains_additional_columns(
        self,
    ):
        """
        A select with a choice filter and no translations in which the first option label is dynamic should maintain data columns.
        """
        xform_md = """
            | survey |                    |      |            |               |         |
            |        | type               | name | label      | choice_filter | default |
            |        | text               | txt  | Enter text |               | default |
            |        | select_one choices | one  | Select one | 1 < 2         |         |
            | choices |
            |         | list_name | name | label        | foo |
            |         | choices   | one  | One - ${txt} | baz |
            """
        self.assertPyxformXform(
            name="data", md=xform_md, model__contains=["<foo>baz</foo>"],
        )

    def test_select_with_dynamic_option_label__and_no_choice_filter__and_no_translations__inlines_output(
        self,
    ):
        """
        A select without a choice filter and no translations in which the first option label is dynamic should not use itext.
        """
        xform_md = """
            | survey |                    |      |            |
            |        | type               | name | label      |
            |        | text               | txt  | Text       |
            |        | select_one choices | one  | Select one |
            | choices |
            |         | list_name | name | label        |
            |         | choices   | one  | One - ${txt} |
            """
        self.assertPyxformXform(
            name="data",
            md=xform_md,
            xml__contains=['<label> One - <output value=" /data/txt "/> </label>'],
        )



# -*- coding: utf-8 -*-
"""
Survey module with XForm Survey objects and utility functions.
"""
from __future__ import print_function

import codecs
import os
import re
import tempfile
import xml.etree.ElementTree as ETree
from collections import defaultdict
from datetime import datetime

from pyxform import constants
from pyxform.utils import (
    BRACKETED_TAG_REGEX,
    LAST_SAVED_REGEX,
    LAST_SAVED_INSTANCE_NAME,
    has_dynamic_label,
)
from pyxform.errors import PyXFormError, ValidationError
from pyxform.external_instance import ExternalInstance
from pyxform.instance import SurveyInstance
from pyxform.instance_info import InstanceInfo
from pyxform.question import Question
from pyxform.section import Section
from pyxform.survey_element import SurveyElement
from pyxform.utils import (
    NSMAP,
    PatchedText,
    basestring,
    get_languages_with_bad_tags,
    node,
    unicode,
)
from pyxform.validators import enketo_validate, odk_validate

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache


RE_PULLDATA = re.compile(r"(pulldata\s*\(\s*)(.*?),")


def register_nsmap():
    """Function to register NSMAP namespaces with ETree"""
    for prefix, uri in NSMAP.items():
        prefix_no_xmlns = prefix.replace("xmlns", "").replace(":", "")
        ETree.register_namespace(prefix_no_xmlns, uri)


register_nsmap()


@lru_cache(maxsize=None)
def is_parent_a_repeat(survey, xpath):
    """
    Returns the XPATH of the first repeat of the given xpath in the survey,
    otherwise False will be returned.
    """
    parent_xpath = "/".join(xpath.split("/")[:-1])
    if not parent_xpath:
        return False

    if survey.any_repeat(parent_xpath):
        return parent_xpath

    return is_parent_a_repeat(survey, parent_xpath)


@lru_cache(maxsize=None)
def share_same_repeat_parent(survey, xpath, context_xpath, reference_parent=False):
    """
    Returns a tuple of the number of steps from the context xpath to the shared
    repeat parent and the xpath to the target xpath from the shared repeat
    parent.

    For example,
        xpath = /data/repeat_a/group_a/name
        context_xpath = /data/repeat_a/group_b/age

        returns (2, '/group_a/name')'
    """

    def _get_steps_and_target_xpath(context_parent, xpath_parent, include_parent=False):
        parts = []
        steps = 1
        if not include_parent:
            remainder_xpath = xpath[len(xpath_parent) :]
            context_parts = context_xpath[len(xpath_parent) + 1 :].split("/")
            xpath_parts = xpath[len(xpath_parent) + 1 :].split("/")
        else:
            split_idx = len(xpath_parent.split("/"))
            context_parts = context_xpath.split("/")[split_idx - 1 :]
            xpath_parts = xpath.split("/")[split_idx - 1 :]
            remainder_xpath = "/".join(xpath_parts)

        for index, item in enumerate(context_parts[:-1]):
            try:
                if xpath[len(context_parent) + 1 :].split("/")[index] != item:
                    steps = len(context_parts[index:])
                    parts = xpath_parts[index:]
                    break
                else:
                    parts = remainder_xpath.split("/")[index + 2 :]
            except IndexError:
                steps = len(context_parts[index - 1 :])
                parts = xpath_parts[index - 1 :]
                break
        return (steps, "/" + "/".join(parts) if parts else remainder_xpath)

    context_parent = is_parent_a_repeat(survey, context_xpath)
    xpath_parent = is_parent_a_repeat(survey, xpath)
    if context_parent and xpath_parent and xpath_parent in context_parent:
        if (not context_parent == xpath_parent and reference_parent) or bool(
            is_parent_a_repeat(survey, context_parent)
        ):
            context_shared_ancestor = is_parent_a_repeat(survey, context_parent)
            if context_shared_ancestor == xpath_parent:
                # Check if context_parent is a child repeat of the xpath_parent
                # If the context_parent is a child of the xpath_parent reference the entire
                # xpath_parent in the generated nodeset
                context_parent = context_shared_ancestor
            elif context_parent == xpath_parent and context_shared_ancestor:
                # If the context_parent is a child of another
                # repeat and is equal to the xpath_parent
                # we avoid refrencing the context_parent and instead reference the shared
                # ancestor
                reference_parent = False
        return _get_steps_and_target_xpath(
            context_parent, xpath_parent, reference_parent
        )
    elif context_parent and xpath_parent:
        # Check if context_parent and xpath_parent share a common
        # repeat ancestor
        context_shared_ancestor = is_parent_a_repeat(survey, context_parent)
        xpath_shared_ancestor = is_parent_a_repeat(survey, xpath_parent)

        if (
            xpath_shared_ancestor
            and context_shared_ancestor
            and xpath_shared_ancestor == context_shared_ancestor
        ):
            return _get_steps_and_target_xpath(
                context_shared_ancestor, xpath_shared_ancestor
            )

    return (None, None)


class Survey(Section):
    """
    Survey class - represents the full XForm XML.
    """

    FIELDS = Section.FIELDS.copy()
    FIELDS.update(
        {
            "_xpath": dict,
            "_created": datetime.now,  # This can't be dumped to json
            "setvalues_by_triggering_ref": dict,
            "title": unicode,
            "id_string": unicode,
            "sms_keyword": unicode,
            "sms_separator": unicode,
            "sms_allow_media": bool,
            "sms_date_format": unicode,
            "sms_datetime_format": unicode,
            "sms_response": unicode,
            constants.COMPACT_PREFIX: unicode,
            constants.COMPACT_DELIMITER: unicode,
            "file_name": unicode,
            "default_language": unicode,
            "_translations": dict,
            "submission_url": unicode,
            "auto_send": unicode,
            "auto_delete": unicode,
            "public_key": unicode,
            "instance_xmlns": unicode,
            "version": unicode,
            "choices": dict,
            "style": unicode,
            "attribute": dict,
            "namespaces": unicode,
        }
    )  # yapf: disable

    def validate(self):
        if self.id_string in [None, "None"]:
            raise PyXFormError("Survey cannot have an empty id_string")
        super(Survey, self).validate()
        self._validate_uniqueness_of_section_names()

    def _validate_uniqueness_of_section_names(self):
        root_node_name = self.name
        section_names = []
        for element in self.iter_descendants():
            if isinstance(element, Section):
                if element.name in section_names:
                    if element.name == root_node_name:
                        # The root node name is rarely explictly set; explain
                        # the problem in a more helpful way (#510)
                        raise PyXFormError(
                            'The name "%s" is the same as the form name. '
                            "Use a different section name "
                            '(or change the form name in the "name" column of '
                            "the settings sheet)." % element.name
                        )
                    raise PyXFormError(
                        "There are two sections with the name %s." % element.name
                    )
                section_names.append(element.name)

    def get_nsmap(self):
        """Add additional namespaces"""
        namespaces = getattr(self, constants.NAMESPACES, None)

        if namespaces and isinstance(namespaces, basestring):
            nslist = [
                ns.split("=")
                for ns in namespaces.split()
                if len(ns.split("=")) == 2 and ns.split("=")[0] != ""
            ]
            xmlns = "xmlns:"
            nsmap = NSMAP.copy()
            nsmap.update(
                dict(
                    [
                        (xmlns + k, v.replace('"', "").replace("'", ""))
                        for k, v in nslist
                        if xmlns + k not in nsmap
                    ]
                )
            )
            return nsmap

        return NSMAP

    def xml(self):
        """
        calls necessary preparation methods, then returns the xml.
        """
        self.validate()
        self._setup_xpath_dictionary()

        for triggering_reference in self.setvalues_by_triggering_ref.keys():
            if not (re.match(BRACKETED_TAG_REGEX, triggering_reference)):
                raise PyXFormError(
                    "Only references to other fields are allowed in the 'trigger' column."
                )

            # try to resolve reference and fail if can't
            self.insert_xpaths(triggering_reference, self)

        body_kwargs = {}
        if hasattr(self, constants.STYLE) and getattr(self, constants.STYLE):
            body_kwargs["class"] = getattr(self, constants.STYLE)
        nsmap = self.get_nsmap()

        return node(
            "h:html",
            node("h:head", node("h:title", self.title), self.xml_model()),
            node("h:body", *self.xml_control(), **body_kwargs),
            **nsmap
        )

    def get_setvalues_for_question_name(self, question_name):
        return self.setvalues_by_triggering_ref.get("${%s}" % question_name)

    @staticmethod
    def _generate_static_instances(list_name, choice_list):
        """
        Generates <instance> elements for static data
        (e.g. choices for select type questions)

        Note that per commit message 0578242 and in xls2json.py R539, an
        instance is only output for select items defined in the choices sheet
        when the item has a choice_filter, and it is that way for backwards
        compatibility.
        """
        instance_element_list = []
        multi_language = isinstance(choice_list[0].get("label"), dict)
        has_media = bool(choice_list[0].get("media"))

        for idx, choice in enumerate(choice_list):
            choice_element_list = []
            # Add a unique id to the choice element in case there is itext
            # it references
            if (
                multi_language
                or has_media
                or has_dynamic_label(choice_list, multi_language)
            ):
                itext_id = "-".join([list_name, str(idx)])
                choice_element_list.append(node("itextId", itext_id))

            for name, value in sorted(choice.items()):
                if isinstance(value, basestring) and name != "label":
                    choice_element_list.append(node(name, unicode(value)))
                if (
                    not multi_language
                    and not has_media
                    and not has_dynamic_label(choice_list, multi_language)
                    and isinstance(value, basestring)
                    and name == "label"
                ):
                    choice_element_list.append(node(name, unicode(value)))

            instance_element_list.append(node("item", *choice_element_list))

        return InstanceInfo(
            type="choice",
            context="survey",
            name=list_name,
            src=None,
            instance=node(
                "instance", node("root", *instance_element_list), id=list_name
            ),
        )

    @staticmethod
    def _generate_external_instances(element):
        if isinstance(element, ExternalInstance):
            name = element["name"]
            extension = element["type"].split("-")[0]
            prefix = "file-csv" if extension == "csv" else "file"
            src = "jr://{}/{}.{}".format(prefix, name, extension)
            return InstanceInfo(
                type="external",
                context="[type: {t}, name: {n}]".format(
                    t=element["parent"]["type"], n=element["parent"]["name"]
                ),
                name=name,
                src=src,
                instance=node("instance", id=name, src=src),
            )

        return None

    @staticmethod
    def _validate_external_instances(instances):
        """
        Must have unique names.

        - Duplications could come from across groups; this checks the form.
        - Errors are pooled together into a (hopefully) helpful message.
        """
        seen = {}
        for i in instances:
            element = i.name
            if seen.get(element) is None:
                seen[element] = [i]
            else:
                seen[element].append(i)
        errors = []
        for element, copies in seen.items():
            if len(copies) > 1:
                contexts = ", ".join(x.context for x in copies)
                errors.append(
                    "Instance names must be unique within a form. "
                    "The name '{i}' was found {c} time(s), "
                    "under these contexts: {contexts}".format(
                        i=element, c=len(copies), contexts=contexts
                    )
                )
        if errors:
            raise ValidationError("\n".join(errors))

    @staticmethod
    def _generate_pulldata_instances(element):
        def get_pulldata_functions(element):
            """
            Returns a list of different pulldata(... function strings if
            pulldata function is defined at least once for any of:
            calculate, constraint, readonly, required, relevant

            :param: element (pyxform.survey.Survey):
            """
            functions_present = []
            for formula_name in constants.EXTERNAL_INSTANCES:
                if "pulldata(" in unicode(element["bind"].get(formula_name)):
                    functions_present.append(element["bind"][formula_name])
            if "pulldata(" in unicode(element["choice_filter"]):
                functions_present.append(element["choice_filter"])
            if "pulldata(" in unicode(element["default"]):
                functions_present.append(element["default"])

            return functions_present

        def get_instance_info(element, file_id):
            uri = "jr://file-csv/{}.csv".format(file_id)

            return InstanceInfo(
                type=u"pulldata",
                context="[type: {t}, name: {n}]".format(
                    t=element[u"parent"][u"type"], n=element[u"parent"][u"name"]
                ),
                name=file_id,
                src=uri,
                instance=node("instance", id=file_id, src=uri),
            )

        pulldata_usages = get_pulldata_functions(element)
        if len(pulldata_usages) > 0:
            pulldata_instances = []
            for usage in pulldata_usages:
                for call_match in re.finditer(RE_PULLDATA, usage):
                    groups = call_match.groups()
                    if len(groups) == 2:
                        first_argument = (  # first argument to pulldata()
                            groups[1].replace("'", "").replace('"', "").strip()
                        )
                        pulldata_instances.append(
                            get_instance_info(element, first_argument)
                        )
            return pulldata_instances
        return None

    @staticmethod
    def _generate_from_file_instances(element):
        itemset = element.get("itemset")
        if itemset and (itemset.endswith(".csv") or itemset.endswith(".xml")):
            file_id, ext = os.path.splitext(itemset)
            uri = "jr://%s/%s" % (
                "file" if ext == ".xml" else "file-%s" % ext[1:],
                itemset,
            )
            return InstanceInfo(
                type="file",
                context="[type: {t}, name: {n}]".format(
                    t=element["parent"]["type"], n=element["parent"]["name"]
                ),
                name=file_id,
                src=uri,
                instance=node("instance", id=file_id, src=uri),
            )

        return None

    # True if a last-saved instance should be generated, false otherwise
    @staticmethod
    def _generate_last_saved_instance(element):
        for expression_type in constants.EXTERNAL_INSTANCES:
            last_saved_expression = re.search(
                LAST_SAVED_REGEX, unicode(element["bind"].get(expression_type))
            )
            if last_saved_expression:
                return True

        return re.search(
            LAST_SAVED_REGEX, unicode(element["choice_filter"])
        ) or re.search(LAST_SAVED_REGEX, unicode(element["default"]))

    @staticmethod
    def _get_last_saved_instance():
        name = (
            "__last-saved"  # double underscore used to minimize risk of name conflicts
        )
        uri = "jr://instance/last-saved"

        return InstanceInfo(
            type="instance",
            context=None,
            name=name,
            src=uri,
            instance=node("instance", id=name, src=uri),
        )

    def _generate_instances(self):
        """
        Get instances from all the different ways that they may be generated.

        An opportunity to validate instances before output to the XML model.

        Instance names used for the id attribute are generated as follows:

        - xml-external: item name value (for type==xml-external)
        - pulldata: first arg to calculation->pulldata()
        - select from file: file name arg to type->itemset
        - choices: list_name (for type==select_*)
        - last-saved: static name of jr://instance/last-saved

        Validation and business rules for output of instances:

        - xml-external item name must be unique across the XForm and the form
          is considered invalid if there is a duplicate name. This differs from
          other item types which allow duplicates if not in the same group.
        - for all instance sources, if the same instance name is encountered,
          the following rules are used to allow re-using instances but prevent
          overwriting conflicting instances:
          - same id, same src URI: skip adding the second (duplicate) instance
          - same id, different src URI: raise an error
          - otherwise: output the instance

        There are two other things currently supported by pyxform that involve
        external files and are not explicitly handled here, but may be relevant
        to future efforts to harmonise / simplify external data workflows:

        - `search` appearance/function: works a lot like pulldata but the csv
          isn't made explicit in the form.
        - `select_one_external`: implicitly relies on a `itemsets.csv` file and
          uses XPath-like expressions for querying.
        """
        instances = []
        generate_last_saved = False
        for i in self.iter_descendants():
            i_ext = self._generate_external_instances(element=i)
            i_pull = self._generate_pulldata_instances(element=i)
            i_file = self._generate_from_file_instances(element=i)
            if not generate_last_saved:
                generate_last_saved = self._generate_last_saved_instance(element=i)
            for x in [i_ext, i_pull, i_file]:
                if x is not None:
                    instances += x if isinstance(x, list) else [x]

        if generate_last_saved:
            instances += [self._get_last_saved_instance()]

        # Append last so the choice instance is excluded on a name clash.
        for name, value in self.choices.items():
            instances += [
                self._generate_static_instances(list_name=name, choice_list=value)
            ]

        # Check that external instances have unique names.
        if instances:
            ext_only = [x for x in instances if x.type == "external"]
            self._validate_external_instances(instances=ext_only)

        seen = {}
        for i in instances:
            if i.name in seen.keys() and seen[i.name].src != i.src:
                # Instance id exists with different src URI -> error.
                msg = (
                    "The same instance id will be generated for different "
                    "external instance source URIs. Please check the form."
                    " Instance name: '{i}', Existing type: '{e}', "
                    "Existing URI: '{iu}', Duplicate type: '{d}', "
                    "Duplicate URI: '{du}', Duplicate context: '{c}'.".format(
                        i=i.name,
                        iu=seen[i.name].src,
                        e=seen[i.name].type,
                        d=i.type,
                        du=i.src,
                        c=i.context,
                    )
                )
                raise PyXFormError(msg)
            elif i.name in seen.keys() and seen[i.name].src == i.src:
                # Instance id exists with same src URI -> ok, don't duplicate.
                continue
            else:
                # Instance doesn't exist yet -> add it.
                yield i.instance
            seen[i.name] = i

    def xml_model(self):
        """
        Generate the xform <model> element
        """
        self._setup_translations()
        self._setup_media()
        self._add_empty_translations()

        model_kwargs = {}
        model_kwargs["odk:xforms-version"] = constants.CURRENT_XFORMS_VERSION

        model_children = []
        if self._translations:
            model_children.append(self.itext())
        model_children += [node("instance", self.xml_instance())]
        model_children += list(self._generate_instances())
        model_children += self.xml_bindings()
        model_children += self.xml_actions()

        if self.submission_url or self.public_key or self.auto_send or self.auto_delete:
            submission_attrs = dict()
            if self.submission_url:
                submission_attrs["action"] = self.submission_url
                submission_attrs["method"] = "post"
            if self.public_key:
                submission_attrs["base64RsaPublicKey"] = self.public_key
            if self.auto_send:
                submission_attrs["orx:auto-send"] = self.auto_send
            if self.auto_delete:
                submission_attrs["orx:auto-delete"] = self.auto_delete
            submission_node = node("submission", **submission_attrs)
            model_children.insert(0, submission_node)

        return node("model", *model_children, **model_kwargs)

    def xml_instance(self, **kwargs):
        result = Section.xml_instance(self, **kwargs)

        # set these first to prevent overwriting id and version
        for key, value in self.attribute.items():
            result.setAttribute(unicode(key), value)

        result.setAttribute("id", self.id_string)

        # add instance xmlns attribute to the instance node
        if self.instance_xmlns:
            result.setAttribute("xmlns", self.instance_xmlns)

        if self.version:
            result.setAttribute("version", self.version)

        if self.prefix:
            result.setAttribute("odk:prefix", self.prefix)

        if self.delimiter:
            result.setAttribute("odk:delimiter", self.delimiter)

        return result

    def _add_to_nested_dict(self, dicty, path, value):
        if len(path) == 1:
            dicty[path[0]] = value
            return
        if path[0] not in dicty:
            dicty[path[0]] = {}
        self._add_to_nested_dict(dicty[path[0]], path[1:], value)

    def _setup_translations(self):
        """
        set up the self._translations dict which will be referenced in the
        setup media and itext functions
        """

        def _setup_choice_translations(name, choice_value, itext_id):
            for media_type_or_language, value in choice_value.items():  # noqa
                if isinstance(value, dict):
                    for language, val in value.items():
                        self._add_to_nested_dict(
                            self._translations,
                            [language, itext_id, media_type_or_language],
                            val,
                        )
                else:
                    if name == "media":
                        self._add_to_nested_dict(
                            self._translations,
                            [self.default_language, itext_id, media_type_or_language],
                            value,
                        )
                    else:
                        self._add_to_nested_dict(
                            self._translations,
                            [media_type_or_language, itext_id, "long"],
                            value,
                        )

        self._translations = defaultdict(dict)  # pylint: disable=W0201
        for element in self.iter_descendants():
            # Skip creation of translations for choices in filtered selects
            # The creation of these translations is done futher below in this
            # function
            parent = element.get("parent")
            if parent and not parent.get("choice_filter"):
                for d in element.get_translations(self.default_language):
                    translation_path = d["path"]
                    form = "long"

                    if "guidance_hint" in d["path"]:
                        translation_path = d["path"].replace("guidance_hint", "hint")
                        form = "guidance"

                    self._translations[d["lang"]][
                        translation_path
                    ] = self._translations[d["lang"]].get(translation_path, {})

                    self._translations[d["lang"]][translation_path].update(
                        {
                            form: {
                                "text": d["text"],
                                "output_context": d["output_context"],
                            }
                        }
                    )

        # This code sets up translations for choices in filtered selects.
        for list_name, choice_list in self.choices.items():
            multi_language = isinstance(choice_list[0].get("label"), dict)
            has_media = bool(choice_list[0].get("media"))
            if (
                not multi_language
                and not has_media
                and not has_dynamic_label(choice_list, multi_language)
            ):
                continue
            for idx, choice in zip(range(len(choice_list)), choice_list):
                for name, choice_value in choice.items():
                    itext_id = "-".join([list_name, str(idx)])
                    if isinstance(choice_value, dict):
                        _setup_choice_translations(name, choice_value, itext_id)
                    elif name == "label":
                        self._add_to_nested_dict(
                            self._translations,
                            [self.default_language, itext_id, "long"],
                            choice_value,
                        )

    def _add_empty_translations(self):
        """
        Adds translations so that every itext element has the same elements \
        accross every language.
        When translations are not provided "-" will be used.
        This disables any of the default_language fallback functionality.
        """
        paths = {}
        for lang, translation in self._translations.items():
            for path, content in translation.items():
                paths[path] = paths.get(path, set()).union(content.keys())

        for lang, translation in self._translations.items():
            for path, content_types in paths.items():
                if path not in self._translations[lang]:
                    self._translations[lang][path] = {}
                for content_type in content_types:
                    if content_type not in self._translations[lang][path]:
                        self._translations[lang][path][content_type] = "-"

    def _setup_media(self):
        """
        Traverse the survey, find all the media, and put in into the \
        _translations data structure which looks like this:
        {language : {element_xpath : {media_type : media}}}
        It matches the xform nesting order.
        """

        def _set_up_media_translations(media_dict, translation_key):
            # This is probably papering over a real problem, but anyway,
            # in py3, sometimes if an item is on an xform with multiple
            # languages and the item only has media defined in # "default"
            # (e.g. no "image" vs. "image::lang"), the media dict will be
            # nested inside of a dict with key "default", e.g.
            # {"default": {"image": "my_image.jpg"}}
            media_dict_default = media_dict.get("default", None)
            if isinstance(media_dict_default, dict):
                media_dict = media_dict_default

            for media_type, possibly_localized_media in media_dict.items():

                if media_type not in SurveyElement.SUPPORTED_MEDIA:
                    raise PyXFormError("Media type: " + media_type + " not supported")

                if isinstance(possibly_localized_media, dict):
                    # media is localized
                    localized_media = possibly_localized_media
                else:
                    # media is not localized so create a localized version
                    # using the default language
                    localized_media = {self.default_language: possibly_localized_media}

                for language, media in localized_media.items():

                    # Create the required dictionaries in _translations,
                    # then add media as a leaf value:

                    if language not in self._translations:
                        self._translations[language] = {}

                    translations_language = self._translations[language]

                    if translation_key not in translations_language:
                        translations_language[translation_key] = {}

                    translations_trans_key = translations_language[translation_key]

                    if media_type not in translations_trans_key:
                        translations_trans_key[media_type] = {}

                    translations_trans_key[media_type] = media

        if not self._translations:
            self._translations = defaultdict(dict)  # pylint: disable=W0201

        for survey_element in self.iter_descendants():
            # Skip set up of media for choices in filtered selects.
            # Translations for the media content should have been set up
            # in _setup_translations
            parent = survey_element.get("parent")
            if parent and not parent.get("choice_filter"):
                translation_key = survey_element.get_xpath() + ":label"
                media_dict = survey_element.get("media")
                _set_up_media_translations(media_dict, translation_key)

    def itext(self):
        """
        This function creates the survey's itext nodes from _translations
        @see _setup_media _setup_translations
        itext nodes are localized images/audio/video/text
        @see http://code.google.com/p/opendatakit/wiki/XFormDesignGuidelines
        """
        result = []
        for lang, translation in self._translations.items():
            if lang == self.default_language:
                result.append(node("translation", lang=lang, default="true()"))
            else:
                result.append(node("translation", lang=lang))

            for label_name, content in translation.items():
                itext_nodes = []
                label_type = label_name.partition(":")[-1]

                if not isinstance(content, dict):
                    raise Exception()

                for media_type, media_value in content.items():
                    if isinstance(media_value, dict):
                        value, output_inserted = self.insert_output_values(
                            media_value["text"], context=media_value["output_context"]
                        )
                    else:
                        value, output_inserted = self.insert_output_values(media_value)

                    if label_type == "hint":
                        if media_type == "guidance":
                            itext_nodes.append(
                                node(
                                    "value",
                                    value,
                                    form="guidance",
                                    toParseString=output_inserted,
                                )
                            )
                        else:
                            itext_nodes.append(
                                node("value", value, toParseString=output_inserted)
                            )
                        continue

                    if media_type == "long":
                        # I'm ignoring long types for now because I don't know
                        # how they are supposed to work.
                        itext_nodes.append(
                            node("value", value, toParseString=output_inserted)
                        )
                    elif media_type == "image":
                        if value != "-":
                            itext_nodes.append(
                                node(
                                    "value",
                                    "jr://images/" + value,
                                    form=media_type,
                                    toParseString=output_inserted,
                                )
                            )
                    else:
                        if value != "-":
                            itext_nodes.append(
                                node(
                                    "value",
                                    "jr://" + media_type + "/" + value,
                                    form=media_type,
                                    toParseString=output_inserted,
                                )
                            )

                result[-1].appendChild(node("text", *itext_nodes, id=label_name))

        return node("itext", *result)

    def date_stamp(self):
        """Returns a date string with the format of %Y_%m_%d."""
        return self._created.strftime("%Y_%m_%d")

    def _to_ugly_xml(self):
        return '<?xml version="1.0"?>' + self.xml().toxml()

    def _to_pretty_xml(self):
        """
        I want the to_xml method to by default validate the xml we are
        producing.
        """
        # Hacky way of pretty printing xml without adding extra white
        # space to text
        # TODO: check out pyxml
        # http://ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-and-silly-whitespace/
        xml_with_linebreaks = self.xml().toprettyxml(indent="  ")
        text_re = re.compile(r"(>)\n\s*(\s[^<>\s].*?)\n\s*(\s</)", re.DOTALL)
        output_re = re.compile(r"\n.*(<output.*>)\n(\s\s)*")
        pretty_xml = text_re.sub(
            lambda m: "".join(m.group(1, 2, 3)), xml_with_linebreaks
        )
        inline_output = output_re.sub(r"\g<1>", pretty_xml)
        return '<?xml version="1.0"?>\n' + inline_output

    def __repr__(self):
        return self.__unicode__()

    def __unicode__(self):
        return "<pyxform.survey.Survey instance at %s>" % hex(id(self))

    def _setup_xpath_dictionary(self):
        self._xpath = {}  # pylint: disable=attribute-defined-outside-init
        for element in self.iter_descendants():
            if isinstance(element, (Question, Section)):
                if element.name in self._xpath:
                    self._xpath[element.name] = None
                else:
                    self._xpath[element.name] = element.get_xpath()

    def _var_repl_function(
        self, matchobj, context, use_current=False, reference_parent=False
    ):
        """
        Given a dictionary of xpaths, return a function we can use to
        replace ${varname} with the xpath to varname.
        """

        name = matchobj.group(2)
        bracketed_tag_name = "${{{0}}}".format(name)
        last_saved = matchobj.group(1) is not None
        is_indexed_repeat = matchobj.string.find("indexed-repeat(") > -1
        indexed_repeat_regex = re.compile(r"indexed-repeat\([^)]+\)")
        function_args_regex = re.compile(r"\b[^()]+\((.*)\)$")
        instance_regex = re.compile(r"instance\([^)]+.+")
        bracket_regex = re.compile(r"\[([^]]+)\]")

        def _in_secondary_instance_predicate():
            """
            check if ${} expression represented by matchobj
            is in a predicate for a path expression for a secondary instance
            """

            if instance_regex.search(matchobj.string) is not None:
                bracket_regex_match_iter = bracket_regex.finditer(matchobj.string)
                # Check whether current ${varname} is in the correct bracket_regex_match
                for bracket_regex_match in bracket_regex_match_iter:
                    if (
                        matchobj.start() >= bracket_regex_match.start()
                        and matchobj.end() <= bracket_regex_match.end()
                    ):
                        return True
                return False
            return False

        def _relative_path(name):
            """Given name in ${name}, return relative xpath to ${name}."""
            return_path = None
            xpath, context_xpath = self._xpath[name], context.get_xpath()
            # share same root i.e repeat_a from /data/repeat_a/...
            if (
                len(context_xpath.split("/")) > 2
                and xpath.split("/")[2] == context_xpath.split("/")[2]
            ):
                # if context xpath and target xpath fall under the same
                # repeat use relative xpath referencing.
                steps, ref_path = share_same_repeat_parent(
                    self, xpath, context_xpath, reference_parent
                )
                if steps:
                    ref_path = ref_path if ref_path.endswith(name) else "/%s" % name
                    prefix = " current()/" if use_current else " "

                    return_path = prefix + "/".join([".."] * steps) + ref_path + " "

            return return_path

        def _is_return_relative_path():
            """Determine condition to return relative xpath of current ${name}."""
            indexed_repeat_relative_path_args_index = [0, 1, 3, 5]
            current_matchobj = matchobj

            if not last_saved and context:

                if not is_indexed_repeat:
                    return True

                # It is possible to have multiple indexed-repeat in an expression
                indexed_repeats_iter = indexed_repeat_regex.finditer(matchobj.string)
                for indexed_repeat in indexed_repeats_iter:

                    # Make sure current ${name} is in the correct indexed-repeat
                    if current_matchobj.end() > indexed_repeat.end():
                        try:
                            next(indexed_repeats_iter)
                            continue
                        except StopIteration:
                            return True

                    # ${name} outside of indexed-repeat always using relative path
                    if (
                        current_matchobj.end() < indexed_repeat.start()
                        or current_matchobj.start() > indexed_repeat.end()
                    ):
                        return True

                    indexed_repeat_name_index = None
                    indexed_repeat_args = (
                        function_args_regex.match(indexed_repeat.group())
                        .group(1)
                        .split(",")
                    )
                    name_arg = "${{{0}}}".format(name)
                    for idx, arg in enumerate(indexed_repeat_args):
                        if name_arg in arg.strip():
                            indexed_repeat_name_index = idx

                    return (
                        indexed_repeat_name_index is not None
                        and indexed_repeat_name_index
                        not in indexed_repeat_relative_path_args_index
                    )

            return False

        intro = (
            "There has been a problem trying to replace %s with the "
            "XPath to the survey element named '%s'." % (matchobj.group(0), name)
        )
        if name not in self._xpath:
            raise PyXFormError(intro + " There is no survey element with this name.")
        if self._xpath[name] is None:
            raise PyXFormError(
                intro + " There are multiple survey elements" " with this name."
            )

        if _is_return_relative_path():
            if not use_current:
                use_current = _in_secondary_instance_predicate()
            relative_path = _relative_path(name)
            if relative_path:
                return relative_path

        last_saved_prefix = (
            "instance('" + LAST_SAVED_INSTANCE_NAME + "')" if last_saved else ""
        )
        return " " + last_saved_prefix + self._xpath[name] + " "

    def insert_xpaths(self, text, context, use_current=False, reference_parent=False):
        """
        Replace all instances of ${var} with the xpath to var.
        """

        def _var_repl_function(matchobj):
            return self._var_repl_function(
                matchobj, context, use_current, reference_parent
            )

        return re.sub(BRACKETED_TAG_REGEX, _var_repl_function, unicode(text))

    def _var_repl_output_function(self, matchobj, context):
        """
        A regex substitution function that will replace
        ${varname} with an output element that has the xpath to varname.
        """
        return '<output value="' + self._var_repl_function(matchobj, context) + '" />'

    def insert_output_values(self, text, context=None):
        """
        Replace all the ${variables} in text with xpaths.
        Returns that and a boolean indicating if there were any ${variables}
        present.
        """

        def _var_repl_output_function(matchobj):
            return self._var_repl_output_function(matchobj, context)

        # There was a bug where escaping is completely turned off in labels
        # where variable replacement is used.
        # For exampke, `${name} < 3` causes an error but `< 3` does not.
        # This is my hacky fix for it, which does string escaping prior to
        # variable replacement:
        text_node = PatchedText()
        text_node.data = text
        xml_text = text_node.toxml()

        # need to make sure we have reason to replace
        # since at this point < is &lt,
        # the net effect &lt gets translated again to &amp;lt;
        if unicode(xml_text).find("{") != -1:
            result = re.sub(
                BRACKETED_TAG_REGEX, _var_repl_output_function, unicode(xml_text)
            )
            return result, not result == xml_text
        return text, False

    # pylint: disable=too-many-arguments
    def print_xform_to_file(
        self, path=None, validate=True, pretty_print=True, warnings=None, enketo=False
    ):
        """
        Print the xForm to a file and optionally validate it as well by
        throwing exceptions and adding warnings to the warnings array.
        """
        if warnings is None:
            warnings = []
        if not path:
            path = self._print_name + ".xml"
        try:
            with codecs.open(path, mode="w", encoding="utf-8") as file_obj:
                if pretty_print:
                    file_obj.write(self._to_pretty_xml())
                else:
                    file_obj.write(self._to_ugly_xml())
        except Exception as error:
            if os.path.exists(path):
                os.unlink(path)
            raise error
        if validate:
            warnings.extend(odk_validate.check_xform(path))
        if enketo:
            warnings.extend(enketo_validate.check_xform(path))

        # Warn if one or more translation is missing a valid IANA subtag
        translations = self._translations.keys()
        if translations:
            bad_languages = get_languages_with_bad_tags(translations)
            if bad_languages:
                warnings.append(
                    "\tThe following language declarations do not contain "
                    "valid machine-readable codes: "
                    + ", ".join(bad_languages)
                    + ". "
                    + "Learn more: http://xlsform.org#multiple-language-support"
                )

    def to_xml(self, validate=True, pretty_print=True, warnings=None, enketo=False):
        """
        Generates the XForm XML.
        validate is True by default - pass the XForm XML through ODK Validator.
        pretty_print is True by default - formats the XML for readability.
        warnings - if a list is passed it stores all warnings generated
        enketo - pass the XForm XML though Enketo Validator.

        Return XForm XML string.
        """
        # On Windows, NamedTemporaryFile must be opened exclusively.
        # So it must be explicitly created, opened, closed, and removed.
        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.close()
        try:
            # this will throw an exception if the xml is not valid
            self.print_xform_to_file(
                path=tmp.name,
                validate=validate,
                pretty_print=pretty_print,
                warnings=warnings,
                enketo=enketo,
            )
        finally:
            if os.path.exists(tmp.name):
                os.remove(tmp.name)
        if pretty_print:
            return self._to_pretty_xml()

        return self._to_ugly_xml()

    def instantiate(self):
        """
        Instantiate as in return a instance of SurveyInstance for collected
        data.
        """
        return SurveyInstance(self)



# -*- coding: utf-8 -*-
"""
Test pyxform.validators.utils module.
"""
import os

from unittest2 import TestCase

from pyxform.tests.utils import prep_class_config
from pyxform.validators.error_cleaner import ErrorCleaner
from pyxform.validators.util import XFORM_SPEC_PATH, check_readable


class TestValidatorUtil(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        prep_class_config(cls=cls)

    def test_cleanup_error_message(self):
        test_str = self.config.get(self.cls_name, "test_cleanup_error_message_test")
        expected_str = self.config.get(
            self.cls_name, "test_cleanup_error_message_expected"
        )
        self.assertEqual(ErrorCleaner.odk_validate(test_str), expected_str.strip())

    def test_do_not_over_trim_javarosa_errors(self):
        test_str = self.config.get(
            self.cls_name, "test_do_not_over_trim_javarosa_errors_test"
        )
        # Unescape tabs in string
        test_str = test_str.replace("\n\\t", "\n\t")
        expected_str = self.config.get(
            self.cls_name, "test_do_not_over_trim_javarosa_errors_expected"
        )
        self.assertEqual(ErrorCleaner.odk_validate(test_str), expected_str.strip())

    def test_single_line_error_still_output(self):
        """Should emit errors that are a single line of text."""
        test_str = self.config.get(
            self.cls_name, "test_single_line_error_still_output_test"
        )
        expected_str = self.config.get(
            self.cls_name, "test_single_line_error_still_output_expected"
        )
        self.assertEqual(ErrorCleaner.odk_validate(test_str), expected_str.strip())

    def test_jarfile_error_returned_asis(self):
        """Should return a jarfile error as-is, to avoid tokenising the path."""
        test_str = self.config.get(
            self.cls_name, "test_jarfile_error_returned_asis_test"
        )
        expected_str = self.config.get(
            self.cls_name, "test_jarfile_error_returned_asis_expected"
        )
        self.assertEqual(ErrorCleaner.odk_validate(test_str), expected_str.strip())


class TestCheckReadable(TestCase):
    def test_check_readable__real_file_ok(self):
        """Should return True if the file is readable."""
        self.assertTrue(check_readable(file_path=XFORM_SPEC_PATH))

    def test_check_readable__fake_file_raises(self):
        """Should raise an error if the file isn't readable."""
        fake_file = os.path.join(os.path.dirname(XFORM_SPEC_PATH), ".fake")
        with self.assertRaises(IOError):
            check_readable(file_path=fake_file, retry_limit=2, wait_seconds=0.1)


class TestErrorMessageCleaning(TestCase):
    def should_clean_odk_validate_stacktrace(self):
        message = """java.lang.NullPointerException Null Pointer\norg.javarosa.xform.parse.XFormParseException Parser"""
        cleaned = ErrorCleaner.odk_validate(message)
        self.assertEqual(cleaned, " Null Pointer\n Parser")

    def should_clean_enketo_validate_stacktrace(self):
        message = """Error in gugu/gaga\nError in gugu/gaga"""
        cleaned = ErrorCleaner.enketo_validate(message)
        self.assertEqual(cleaned, "Error in gugu/gaga")
